#if  defined  (HAVE_CONFIG_H)
#  include "config.h"
#endif

#ifndef _AFTERDOC_H
#  include "afterdoc.h"
#endif
#ifndef _AFTER_H
#  include "after.h"
#endif
#ifndef _CONSTANTS_H
#  include "constants.h"
#endif
#ifndef _VINTERP_H
#  include "vinterp.h"
#endif
#ifndef _CDI_H
#  include "cdi.h"
#endif

#if  defined  (HAVE_LIBPTHREAD)
#  include <pthread.h>
#endif

#if defined (_OPENMP)
#  include <omp.h>
#endif

#if ! defined (VERSION)
#  define  VERSION  "0.0.1"
#endif


void gmInqHistory(int fileID);
void gmDefHistory(int fileID, char *histstring);

char   *zaxisNamePtr(int leveltype);
char   *vlistInqVarNamePtr(int vlistID, int varID);
char   *vlistInqVarLongnamePtr(int vlistID, int varID);
char   *vlistInqVarUnitsPtr(int vlistID, int varID);

typedef struct {
  int lana, nrecs;
}
RARG;

extern double VCT_L191[];

int stdin_is_tty  = 0;  /* true if stdin  is character device */
int stdout_is_tty = 0;  /* true if stdout is character device */

int Source = 0;

int FieldDim = 0;

static void Usage(void);

#define MaxHours  24

static int ifiletype = -1;
static int ofiletype = -1;

static int DataType = -1;

static char *Vctfile = NULL;
static char *filename;
static char **ifiles;
static char *ifile  = NULL;
static char *ofile  = NULL;
static char *ofile2 = NULL;

static int taxisID  = -1;
static int taxisID2 = -1;
static int specGridID  = -1;
static int gaussGridID = -1;
static int iVertID = -1;
static int oVertID = -1;

static int Lhybrid2pressure = FALSE;


int    AnalysisData; /* 0 = ECHAM Data, 1 = ECMWF Spectral Analyses */
int    DayIn       ; /* day increment of infiles if Multi = TRUE    */
int    Debug       = FALSE;
int    Extrapolate = TRUE;
int    Szip        = FALSE;

int    Dim3FC,    Dim3SP,    Dim3GP;
int    DimFC,     DimGP,     DimSP;
int    Dim3FCOut, Dim3SPOut, Dim3GPOut;
int    DimFCOut,  DimGPOut,  DimSPOut;
int    DimSPOut_half ;

int    TsID;
#if  defined  (HAVE_LIBPTHREAD)
int    ParallelRead = TRUE;
#else
int    ParallelRead = FALSE;
#endif
int    ECMWFMars    = FALSE;

#define TIMESTEP_INTERVAL  -1
#define MONTHLY_INTERVAL    0
#define DAILY_INTERVAL      1
#define UNLIM_INTERVAL      2

int    OutputInterval;
int    EndOfInterval ;

int    Fouriers;
int    FouriersOut;
int    Gaussian = FALSE;
int    HalfLevels;
int    Latitudes = 0;
int    LatitudesOut;
int    Longitudes = 0;
int    LongitudesOut;
int    mars;

int    Mean;
int    MeanCount0;
int    MeanCount;
int    Multi;
int    Nfiles = 0;
int    Spectral = FALSE;
int    TermCount;
int    Truncation = 0;
int    TruncationOut;
int    Waves;
int    WavesOut;

int    labort = TRUE;

int instID, modelID;

struct Date NextDate;
struct Date NewDate;
struct Date OldDate;
struct Date StartDate;


struct Control All[MaxCodes+5];


int    *vert_index = NULL;
int    *pnmiss = NULL;
double *dv2uv_f1;
double *dv2uv_f2;
double *Orography;
double *poli = NULL;
double *poliOut = NULL;
double *pold = NULL;
double *pdev = NULL;
double *pol2 = NULL;
double *pol3 = NULL;
double *p_of_height = NULL;

int nrqh;

int nvct = 0;
double *vct = NULL;

int experimentID;

static int Type;
int unitsel;
int hours[MaxHours+1];

static double  LevelRequest[MaxLevel];
static double *LevelFound;

static int NumLevel        = 0;
static int NumLevelFound   = 0;
int NumLevelRequest = 0;
static int NumCodesRequest = 0;

double *rcoslat          = NULL;
double *coslat           = NULL;
double *rcoslatOut       = NULL;
double *coslatOut        = NULL;
double *DerivationFactor = NULL;
double *Field            = NULL;

static int istreamID   = CDI_UNDEFID;
static int ostreamID   = CDI_UNDEFID;
static int ostreamID2  = CDI_UNDEFID;
static int ivlistID    = CDI_UNDEFID;
static int ovlistID    = CDI_UNDEFID;
static int ovlistID2   = CDI_UNDEFID;

void cdiError(int cdiErrno, const char *fmt, ...)
{
  va_list args;
	
  va_start(args, fmt);

  printf("\n");
  vfprintf(stderr, fmt, args);
   fprintf(stderr, "\n");

  va_end(args);

  fprintf(stderr, "%s\n", cdiStringError(cdiErrno));

  if ( _ExitOnError ) exit(1);
}

void
cfprintf(FILE *fp, int num, int cval)
{
  int inum;
  for (inum = 0; inum < num; inum++)
    fprintf(fp, "%c", cval);
  fprintf(fp, "\n");
}


void lprintf(FILE *fp)
{
  int inum;
  int num = 67;
  int cval = '-';

  fprintf(fp, " ");
  for (inum = 0; inum < num; inum++)
    fprintf(fp, "%c", cval);
  fprintf(fp, "\n");
}


char *FieldName(int code, char *Text)
{
  static char AllocName[256];

  sprintf(AllocName, "[%3d].%s", code, Text);

  return (AllocName);
}


char *FieldName2(int code, char *Text)
{
  static char AllocName[256];

  sprintf(AllocName, "[%3d].%s", code, Text);

  return (AllocName);
}

static
void GP2FC(double *gp, double *fc, long nlat, long nlon, long nlev, long nfc)
{
  static char func[] = "GP2FC";
  static long ifax[10];
  static double *trig = NULL;

  if ( ifax[9] != nlon )
    {
      if ( trig ) free (trig);
      trig = (double *) malloc(nlon * sizeof(double));
      fft_set (trig, ifax, nlon);
    }

  gp2fc(trig, ifax, gp, fc, nlat, nlon, nlev, nfc);
}

static
void FC2GP(double *fc, double *gp, long nlat, long nlon, long nlev, long nfc)
{
  static char func[] = "FC2GP";
  static long ifax[10];
  static double *trig = NULL;

  if ( ifax[9] != nlon )
    {
      if ( trig ) free (trig);
      trig = (double *) malloc(nlon * sizeof(double));
      fft_set (trig, ifax, nlon);
    }

  fc2gp(trig, ifax, fc, gp, nlat, nlon, nlev, nfc);
}

static
void legini(int ntr, int nlat)
{
  int waves, dimsp;
  int jgl, jm, jn;
  int jsp, pdim;
  double *gmu, *gwt, *hnm, *pnm, gmusq, *ztemp1, *ztemp2;

  waves =  ntr + 1;
  dimsp = (ntr + 1) * (ntr + 2);
  pdim  = dimsp / 2 * nlat;
  poli  = DoubleAlloc(pdim, "poli.permanent");
  if ( ! AnalysisData )
    {
      if ( Type >= 20 )   pold = DoubleAlloc(pdim, "pold.permanent");
      if ( All[DPSDY].needed ) pdev = DoubleAlloc(pdim, "pdev.permanent");
    }
  if ( (All[DIVERGENCE].needed || All[VORTICITY].needed ||
	All[VELOPOT].needed || All[STREAM].needed ) && Type > 20 )
    {
      pol2 = DoubleAlloc(pdim, "pol2.permanent");
      pol3 = DoubleAlloc(pdim, "pol3.permanent");
    }
  gmu    = DoubleAlloc(nlat,  "legini.gmu");
  gwt    = DoubleAlloc(nlat,  "legini.gwt");

  gaussaw(gmu, gwt, nlat);

  if (Debug)
    {
      fprintf(stderr,"+    legini: Latitudes\n");
      fprintf(stderr,"+           ");
      for (jgl = 0; jgl < nlat; jgl++)
	{
          fprintf(stderr,"%7.2f",180.*asin(gmu[jgl])/M_PI);
          if ((jgl+1)%8 == 0) fprintf(stderr,"\n+           ");
	}
      fprintf(stderr,"\n");
    }

  if (AnalysisData && (Type == 70) && Gaussian && !Spectral)
    {
      if (poli) poli = FreeMemory(poli);
      if (pol2) pol2 = FreeMemory(pol2);
      if (pol3) pol3 = FreeMemory(pol3);
      return;
    }

#if ! defined (_OPENMP)
  hnm    = DoubleAlloc(dimsp, "legini.hnm");
  pnm    = DoubleAlloc(dimsp, "legini.pnm");
  ztemp1 = DoubleAlloc(waves<<1, "ztemp1");
  ztemp2 = DoubleAlloc(waves<<1, "ztemp2");
#endif

#if defined (_OPENMP)
#pragma omp parallel for default(shared) private(jm, jn, jsp, gmusq, hnm, pnm, ztemp1, ztemp2)
#endif
  for ( jgl = 0; jgl < nlat; jgl++ )
    {
#if defined (_OPENMP)
      hnm    = DoubleAlloc(dimsp, "legini.hnm");
      pnm    = DoubleAlloc(dimsp, "legini.pnm");
      ztemp1 = DoubleAlloc(waves<<1, "ztemp1");
      ztemp2 = DoubleAlloc(waves<<1, "ztemp2");
#endif
      gmusq = 1.0 - gmu[jgl]*gmu[jgl];
      coslat[jgl] =  sqrt(gmusq);
      rcoslat[jgl] = 1.0 / coslat[jgl];
      DerivationFactor[jgl] = rcoslat[jgl] / PlanetRadius;

      phcs(pnm, hnm, waves, gmu[jgl], ztemp1, ztemp2);

      jsp = jgl;
      for ( jm = 0; jm < waves; jm++ )
	for ( jn = 0; jn < waves - jm; jn++ )
	  {
	              poli[jsp] = pnm[jm*waves+jn] * 2.0;
	    if (pold) pold[jsp] = pnm[jm*waves+jn] * gwt[jgl];
	    if (pdev) pdev[jsp] = hnm[jm*waves+jn] * 2.0 /
                                  (-PlanetRadius   * sqrt(gmusq));
	    if (pol2) pol2[jsp] = hnm[jm*waves+jn] * gwt[jgl] /
                                  (PlanetRadius    * gmusq);
	    if (pol3) pol3[jsp] = pnm[jm*waves+jn] * gwt[jgl] * jm /
                                  (PlanetRadius    * gmusq);
	    jsp += nlat;
	  }
#if defined (_OPENMP)
      ztemp2 = FreeMemory(ztemp2);
      ztemp1 = FreeMemory(ztemp1);
      pnm    = FreeMemory(pnm);
      hnm    = FreeMemory(hnm);
#endif
    }

#if ! defined (_OPENMP)
  ztemp2 = FreeMemory(ztemp2);
  ztemp1 = FreeMemory(ztemp1);
  pnm    = FreeMemory(pnm);
  hnm    = FreeMemory(hnm);
#endif
  gwt    = FreeMemory(gwt);
  gmu    = FreeMemory(gmu);
}

static
void leginiOut(int ntr, int nlat)
{
  int waves, dimsp;
  int jgl, jm, jn;
  int jsp, pdim;
  double *gmu, *gwt, *hnm, *pnm, gmusq, *ztemp1, *ztemp2;

  if ( TruncationOut == Truncation )
    {
      poliOut    = poli;
      memcpy(rcoslatOut, rcoslat , nlat*sizeof(double));
      memcpy(coslatOut,   coslat,  nlat*sizeof(double));
      return;
    }

  waves =  ntr + 1;
  dimsp = (ntr + 1) * (ntr + 2);
  pdim  = dimsp /2 * nlat;

  poliOut = DoubleAlloc(pdim, "poliOut.permanent");
  gmu     = DoubleAlloc(nlat, "leginiOut.gmu");
  gwt     = DoubleAlloc(nlat, "leginiOut.gwt");
  hnm     = DoubleAlloc(dimsp, "leginiOut.hnm");
  pnm     = DoubleAlloc(dimsp, "leginiOut.pnm");
  ztemp1  = DoubleAlloc(waves<<1, "ztemp1Out");
  ztemp2  = DoubleAlloc(waves<<1, "ztemp2Out");

  gaussaw(gmu, gwt, nlat);

  for (jgl = 0; jgl < nlat; jgl++)
    {
      gmusq = 1.0 - gmu[jgl] * gmu[jgl];
      coslatOut[jgl] =  sqrt(gmusq);
      rcoslatOut[jgl] = 1.0 / coslatOut[jgl];

      phcs(pnm, hnm, waves, gmu[jgl], ztemp1, ztemp2);

      jsp = jgl;
      for (jm = 0; jm < waves; jm++)
	{
	  for (jn = 0; jn < waves - jm; jn++)
	    {
	      poliOut[jsp] = pnm[jm*waves+jn] * 2.0;
	      jsp += nlat;
	    }
	}
    }

  ztemp2 = FreeMemory(ztemp2);
  ztemp1 = FreeMemory(ztemp1);
  pnm    = FreeMemory(pnm);
  hnm    = FreeMemory(hnm);
  gwt    = FreeMemory(gwt);
  gmu    = FreeMemory(gmu);
}

static
void gp2sp(int ccode)
{
  long fieldSize;

  if ( All[ccode].spectral == NULL )
    {
      if ( All[ccode].hybrid == NULL )
	{
	  fprintf(stderr,"%d.hybrid not found\n", ccode);
	  exit(99);
	}
      if ( All[ccode].fourier == NULL )
	{
	  fieldSize = DimFC * All[ccode].hlev;
	  All[ccode].fourier = DoubleAlloc(fieldSize,"gp2sp.fourier");
	  GP2FC(All[ccode].hybrid,All[ccode].fourier,
		Latitudes,Longitudes,All[ccode].hlev,Fouriers);
	}
      All[ccode].spectral = DoubleAlloc(Dim3SP,"gp2sp.spectral");
      fc2sp(All[ccode].fourier,All[ccode].spectral,
            pold,All[ccode].hlev,Latitudes,Fouriers,Truncation);    
    }
}

static
void SPuv2SPdv()
{
   int i;
   long fieldSize;
   double *Div, *DivOut, *Vor, *VorOut;

   Div = DivOut = All[DIVERGENCE].spectral;
   Vor = VorOut =  All[VORTICITY].spectral;
   fieldSize = DimFC * NumLevelRequest;

   if ( All[U_WIND].fourier == NULL )
     All[U_WIND].fourier = DoubleAlloc (fieldSize, "All[U_WIND].fourier");
   if ( All[V_WIND].fourier == NULL )
     All[V_WIND].fourier = DoubleAlloc (fieldSize, "All[V_WIND].fourier");

   sp2fc(All[U_WIND].spectral, All[U_WIND].fourier, poli,
         NumLevelRequest, Latitudes, Fouriers, Truncation);
   sp2fc(All[V_WIND].spectral, All[V_WIND].fourier, poli,
         NumLevelRequest, Latitudes, Fouriers, Truncation);
   uv2dv(All[U_WIND].fourier, All[V_WIND].fourier, Div, Vor,
         pol2, pol3, NumLevelRequest, Latitudes, Truncation);

   All[U_WIND].fourier = FreeMemory(All[U_WIND].fourier);
   All[V_WIND].fourier = FreeMemory(All[V_WIND].fourier);

   for (i = 0; i < NumLevelRequest; ++i) {
      sp2sp(Div, Truncation, DivOut, TruncationOut);
      sp2sp(Vor, Truncation, VorOut, TruncationOut);
      Div    += DimSP;
      Vor    += DimSP;
      DivOut += DimSPOut;
      VorOut += DimSPOut;
   }
}

/* HUMTEST */
void FCrh2FCsh(void)
{
   long fieldSize;

   fieldSize = DimGPOut * NumLevelRequest;

   if ( All[RHUMIDITY].grid == NULL )
     All[RHUMIDITY].grid = DoubleAlloc (fieldSize, "All[RHUMIDITY].grid");
   if ( All[TEMPERATURE].grid == NULL )
     All[TEMPERATURE].grid = DoubleAlloc (fieldSize, "All[TEMPERATURE].grid");
   if ( All[HUMIDITY].grid == NULL )
     All[HUMIDITY].grid = DoubleAlloc (fieldSize, "All[HUMIDITY].grid");

   FC2GP(All[RHUMIDITY].fourier, All[RHUMIDITY].grid,
	 LatitudesOut, LongitudesOut, All[RHUMIDITY].plev, FouriersOut);
   FC2GP(All[TEMPERATURE].fourier, All[TEMPERATURE].grid,
	 LatitudesOut, LongitudesOut, All[TEMPERATURE].plev, FouriersOut);

   rh2sh(All[HUMIDITY].grid, All[RHUMIDITY].grid, All[TEMPERATURE].grid, NumLevelRequest,
	 DimGPOut, LevelRequest);

   GP2FC(All[HUMIDITY].grid, All[HUMIDITY].fourier,
	 LatitudesOut, LongitudesOut, All[HUMIDITY].plev, FouriersOut);

   All[HUMIDITY].grid    = FreeMemory(All[HUMIDITY].grid);
   All[RHUMIDITY].grid   = FreeMemory(All[RHUMIDITY].grid);
   All[TEMPERATURE].grid = FreeMemory(All[TEMPERATURE].grid);
}

void FCsh2FCrh(void)
{
   long fieldSize;

   fieldSize = DimGPOut * NumLevelRequest;

   if ( All[RHUMIDITY].grid == NULL )
     All[RHUMIDITY].grid = DoubleAlloc (fieldSize, "All[RHUMIDITY].grid");
   if ( All[TEMPERATURE].grid == NULL )
     All[TEMPERATURE].grid = DoubleAlloc (fieldSize, "All[TEMPERATURE].grid");
   if ( All[HUMIDITY].grid == NULL )
     All[HUMIDITY].grid = DoubleAlloc (fieldSize, "All[HUMIDITY].grid");

   FC2GP(All[HUMIDITY].fourier, All[HUMIDITY].grid,
	 LatitudesOut, LongitudesOut, All[HUMIDITY].plev, FouriersOut);
   FC2GP(All[TEMPERATURE].fourier, All[TEMPERATURE].grid,
	 LatitudesOut, LongitudesOut, All[TEMPERATURE].plev, FouriersOut);

   sh2rh(All[HUMIDITY].grid, All[RHUMIDITY].grid, All[TEMPERATURE].grid, NumLevelRequest,
	 DimGPOut, LevelRequest, All[FULL_PRESS].hybrid);

   GP2FC(All[RHUMIDITY].grid, All[RHUMIDITY].fourier,
	 LatitudesOut, LongitudesOut, All[RHUMIDITY].plev, FouriersOut);

   All[HUMIDITY].grid    = FreeMemory(All[HUMIDITY].grid);
   All[RHUMIDITY].grid   = FreeMemory(All[RHUMIDITY].grid);
   All[TEMPERATURE].grid = FreeMemory(All[TEMPERATURE].grid);
}
/* ENDE HUMTEST */


/*
 *  Process Pressure Level data
 */
void processPL(void)
{
  static char func[] = "processPL";
  int code, l;
  long fieldSize;
  int lindex, nlevel;
  int offset;

  MeanCount++;
  TermCount++;

  if ( MeanCount == 1 )
    {
      if ( Debug )
	fprintf(stderr, "CheckAnalyses: %d %d\n", TermCount, MeanCount);
      CheckAnalyses(All);
      StartDate = OldDate;
    }
  if ( TermCount  > 120 ) Debug = 0;

  /* ============================== */
  /* Computations in spectral space */
  /* ============================== */

  if ( All[TEMPERATURE].needed )
    {
      All[TEMPERATURE].hlev = 2;
      All[TEMPERATURE].plev = NumLevelRequest;
      All[TEMPERATURE].sfit = TRUE;
    }

  if ( All[GEOPOTHEIGHT].comp )
    {
      All[GEOPOTHEIGHT].hlev = 2;
      All[GEOPOTHEIGHT].plev = NumLevelRequest;
      All[GEOPOTHEIGHT].sfit = TRUE;
    }

  if ( All[GEOPOTHEIGHT].comp && All[GEOPOTENTIAL].detected )
    {
      if ( All[GEOPOTHEIGHT].spectral == NULL )
	All[GEOPOTHEIGHT].spectral = DoubleAlloc(DimSPOut*NumLevelRequest,
						  "GEOPOTHEIGHT.spectral");
      MultVectorScalar(All[GEOPOTHEIGHT].spectral, All[GEOPOTENTIAL].spectral,
		       C_RG, DimSPOut*NumLevelRequest, 0, 0);
      All[GEOPOTENTIAL].needed = All[GEOPOTENTIAL].selected;
    }

  if ( Type == 50 && All[HUMIDITY].needed && All[HUMIDITY].spectral == NULL )
    {
      All[HUMIDITY].plev = NumLevelRequest;
      All[HUMIDITY].sfit = TRUE;
      All[HUMIDITY].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[HUMIDITY].spectral");
      /*
	SPrh2SPsh();
      */
      All[RHUMIDITY].needed = All[RHUMIDITY].selected;
      All[TEMPERATURE].needed = All[TEMPERATURE].selected;
    }

  if ( All[U_WIND].spectral && All[V_WIND].spectral &&
       (All[DIVERGENCE].comp || All[VORTICITY].comp) )
    {
      All[DIVERGENCE].hlev = All[VORTICITY].hlev =    2;
      All[DIVERGENCE].plev = All[VORTICITY].plev = NumLevelRequest;
      All[DIVERGENCE].sfit = All[VORTICITY].sfit = TRUE;
      if ( All[DIVERGENCE].spectral == NULL )
	All[DIVERGENCE].spectral = DoubleAlloc(DimSP*NumLevelRequest, "All[DIVERGENCE].spectral");
      if ( All[VORTICITY].spectral == NULL )
	All[VORTICITY].spectral = DoubleAlloc(DimSP*NumLevelRequest, "All[VORTICITY].spectral");
      SPuv2SPdv();
    }

  if ( All[U_WIND].comp || All[V_WIND].comp )
    {
      All[U_WIND].hlev = All[V_WIND].hlev =    2;
      All[U_WIND].plev = All[V_WIND].plev = NumLevelRequest;
      All[U_WIND].sfit = All[V_WIND].sfit = TRUE;
      if ( All[U_WIND].spectral == NULL )
	All[U_WIND].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[U_WIND].spectral");
      if ( All[V_WIND].spectral == NULL )
	All[V_WIND].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[V_WIND].spectral");
      dv2uv(All[DIVERGENCE].spectral, All[VORTICITY].spectral,
	    All[U_WIND].spectral, All[V_WIND].spectral,
	    dv2uv_f1, dv2uv_f2,
	    TruncationOut, DimSPOut, NumLevelRequest);
    }

  if ( All[VELOPOT].comp )
    {
      All[VELOPOT].hlev =    2;
      All[VELOPOT].plev = NumLevelRequest;
      All[VELOPOT].sfit = TRUE;
      if ( All[VELOPOT].spectral == NULL )
	All[VELOPOT].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[VELOPOT].spectral");
      dv2ps(All[DIVERGENCE].spectral, All[VELOPOT].spectral, NumLevelRequest, TruncationOut);
    }

  if ( All[STREAM].comp )
    {
      All[STREAM].hlev =    2;
      All[STREAM].plev = NumLevelRequest;
      All[STREAM].sfit = TRUE;
      if ( All[STREAM].spectral == NULL )
	All[STREAM].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[STREAM].spectral");
      dv2ps(All[VORTICITY].spectral, All[STREAM].spectral, NumLevelRequest, TruncationOut);
    }

  /* --------------------------- */
  /*  Output of spectral fields  */
  /* --------------------------- */

  if ( Type == 50 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].spectral )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].spectral, 0);
	    else
	      Error(func, "Code %d not available on spectral space!", code);
	  }

      FreeSpectral(All);
      return;
    }

  /* =============================== */
  /* Transformation to fourier space */
  /* Computations in fourier space   */
  /* =============================== */

  if ( Type >= 60 )
    {
      for (code = 0; code < MaxCodes; code++)
	if (All[code].needed && All[code].spectral)
	  {
	    if (All[code].fourier == NULL)
	      {
		fieldSize = All[code].plev * DimFCOut;
		All[code].fourier = DoubleAlloc(fieldSize, FieldName(code,"fourier"));
	      }
	    sp2fc(All[code].spectral,All[code].fourier,poliOut,
		  All[code].plev,LatitudesOut,FouriersOut,TruncationOut);
	  }
      if ( All[U_WIND].needed && All[U_WIND].fourier )
	scaluv(All[U_WIND].fourier, rcoslatOut, LatitudesOut, FouriersOut*NumLevelRequest);
      if ( All[V_WIND].needed && All[V_WIND].fourier )
	scaluv(All[V_WIND].fourier, rcoslatOut, LatitudesOut, FouriersOut*NumLevelRequest);

      /* HUMTEST */
      if ( Type < 70 && All[HUMIDITY].needed && All[HUMIDITY].fourier == NULL )
	{
	  All[HUMIDITY].plev = NumLevelRequest;
	  All[HUMIDITY].sfit = TRUE;
	  All[HUMIDITY].fourier = DoubleAlloc(DimFCOut*NumLevelRequest, "All[HUMIDITY].fourier");

	  FCrh2FCsh();

	  All[RHUMIDITY].needed = All[RHUMIDITY].selected;
	  All[TEMPERATURE].needed = All[TEMPERATURE].selected;
	}

      if ( Type < 70 && All[RHUMIDITY].needed && All[RHUMIDITY].fourier == NULL )
	{
	  All[RHUMIDITY].plev = NumLevelRequest;
	  All[RHUMIDITY].sfit = TRUE;
	  All[RHUMIDITY].fourier = DoubleAlloc(DimFCOut*NumLevelRequest,
					       "All[RHUMIDITY].fourier");

	  FCsh2FCrh();

	  All[HUMIDITY].needed = All[HUMIDITY].selected;
	  All[TEMPERATURE].needed = All[TEMPERATURE].selected;
	}
      /* ENDE HUMTEST */
    }

  FreeSpectral(All);

  /* -------------------------- */
  /*  Output of fourier fields  */
  /* -------------------------- */

  if ( Type == 60 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].fourier, 0);
	    else
	      Error(func, "Code %d not available on fourier space!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ----------------------- */
  /*  Output of zonal means  */
  /* ----------------------- */

  if ( Type == 61 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      {
		nlevel = zaxisInqSize(All[code].ozaxisID);
		for ( lindex = 0; lindex < nlevel; lindex++ )
		  {
		    offset = lindex*DimFCOut;
		    streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].fourier+offset, 0);
		  }
	      }
	    else
	      Error(func, "Code %d not available on zonal mean!", code);
	  }

      FreeFourier(All);
      return;
   }

  /* ============================ */
  /* Transformation to gridpoints */
  /* ============================ */

  if ( All[PS].comp && All[LNPS].grid )
    {
      if (All[PS].grid == NULL) All[PS].grid = DoubleAlloc(DimGP, "Ps");
      for (l = 0; l < DimGP; l++) All[PS].grid[l] = exp(All[LNPS].grid[l]);
    }

  if ( Type >= 70 )
    {
      for (code = 0; code < MaxCodes; code++)
	if (All[code].needed && All[code].fourier)
	  {
	    if (All[code].grid == NULL)
	      {
		fieldSize = All[code].plev * DimGPOut;
		All[code].grid = DoubleAlloc(fieldSize,FieldName(code,"grid"));
	      }

	    FC2GP(All[code].fourier,All[code].grid,
		  LatitudesOut,LongitudesOut,All[code].plev,FouriersOut);
	  }
    }

   FreeFourier(All);

  /* HUMTEST */
  /* -------------------------------- */
  /*  Computation in gridpoint space  */
  /* -------------------------------- */

  if ( All[RHUMIDITY].needed && All[RHUMIDITY].grid == NULL )
    {
      All[RHUMIDITY].plev = NumLevelRequest;
      All[RHUMIDITY].sfit = TRUE;
      All[RHUMIDITY].grid = DoubleAlloc(DimGPOut*NumLevelRequest, "All[RHUMIDITY].grid");
      sh2rh(All[HUMIDITY].grid, All[RHUMIDITY].grid, All[TEMPERATURE].grid, NumLevelRequest,
	    DimGPOut, LevelRequest, All[FULL_PRESS].hybrid);
      All[HUMIDITY].needed = All[HUMIDITY].selected;
      All[TEMPERATURE].needed = All[TEMPERATURE].selected;
    }

  if ( All[HUMIDITY].needed && All[HUMIDITY].grid == NULL )
    {
      All[HUMIDITY].plev = NumLevelRequest;
      All[HUMIDITY].sfit = TRUE;
      All[HUMIDITY].grid = DoubleAlloc(DimGPOut*NumLevelRequest, "All[HUMIDITY].grid");
      rh2sh(All[HUMIDITY].grid, All[RHUMIDITY].grid, All[TEMPERATURE].grid, NumLevelRequest,
	    DimGPOut, LevelRequest);
      All[RHUMIDITY].needed = All[RHUMIDITY].selected;
      All[TEMPERATURE].needed = All[TEMPERATURE].selected;
    }
  /* HUMTEST ENDE */

  /* -------------------------- */
  /*  Computation of Means      */
  /* -------------------------- */

  if ( Mean )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].needed && All[code].grid )
	{
	  fieldSize = DimGPOut * All[code].plev;
	  if (All[code].mean == NULL)
	    All[code].mean = DoubleAlloc(fieldSize,FieldName(code,"mean"));

	  if (MeanCount == 1) RealCopy(All[code].mean, All[code].grid, fieldSize);
	  else                AddVector(All[code].mean, All[code].grid, fieldSize,
					&All[code].nmiss, All[code].missval);

	  if ( EndOfInterval )
	    {
	      if ( All[code].samp == NULL )
		MultVectorScalar(All[code].mean, All[code].mean, 1.0/MeanCount, fieldSize,
				 All[code].nmiss, All[code].missval);
	      else
		DivVectorIvector(All[code].mean, All[code].mean, All[code].samp, fieldSize,
				 &All[code].nmiss, All[code].missval);
	    }
	}

  /* -------------------------- */
  /*  Computation of Variances  */
  /* -------------------------- */

  if ( Mean > 1 )
    for (code = 0; code < MaxCodes; code++)
      if (All[code].needed && All[code].mean)
	{
	  fieldSize = DimGPOut * All[code].plev;
	  if (All[code].variance == NULL)
	    All[code].variance = DoubleAlloc(fieldSize,FieldName(code,"var"));
	  if (MeanCount == 1) IniQuaSum(All[code].variance,All[code].grid,fieldSize);
	  else                AddQuaSum(All[code].variance,All[code].grid,fieldSize);

	  if (EndOfInterval) VarQuaSum(All[code].variance,All[code].mean,fieldSize, MeanCount);
	}

  if ( Mean && !EndOfInterval )
    {
      FreeGrid(All);
      return;
    }

  /* ---------------------------------------------- */
  /*  Output of pressure level means and variances  */
  /* ---------------------------------------------- */

  if ( Type == 70 && Mean && EndOfInterval )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( Mean != 2 )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].mean, All[code].nmiss);
	    if ( Mean >= 2 )
	      streamWriteVar(ostreamID2, All[code].ovarID2, All[code].variance, All[code].nmiss);
	  }

      FreeSamp(All);
      FreeGrid(All);
      return;
    }

  /* -------------------------------- */
  /*  Output of pressure level grids  */
  /* -------------------------------- */

  if ( Type == 70 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  streamWriteVar(ostreamID, All[code].ovarID, All[code].grid, All[code].nmiss);

      FreeGrid(All);
      return;
    }
}

void EchamCompGP(void)  /* Grid Point Computations */
{
  int l;

  if ( All[GEOPOTHEIGHT].comp || All[SLP].comp || All[THETAF].needed ||
       All[HALF_PRESS].needed || All[RHUMIDITY].comp || All[OMEGA].comp ||
       Type >= 30 )
    {
      if ( All[FULL_PRESS].hybrid == NULL )
	All[FULL_PRESS].hybrid = DoubleAlloc(Dim3GP, "All[FULL_PRESS].hybrid");

      All[HALF_PRESS].hlev = NumLevel+1;
      All[HALF_PRESS].plev = NumLevelRequest;
      All[HALF_PRESS].sfit = FALSE;

      if ( All[HALF_PRESS].hybrid == NULL )
	All[HALF_PRESS].hybrid = DoubleAlloc(Dim3GP+DimGP, "All[HALF_PRESS].hybrid");

      presh(All[FULL_PRESS].hybrid, All[HALF_PRESS].hybrid, vct, All[PS_PROG].hybrid,
	    NumLevel, DimGP);
    }

  if ( unitsel > 2 ) All[FULL_PRESS].hybrid = FreeMemory(All[FULL_PRESS].hybrid);

  if ( All[THETAF].needed )
    {
      All[THETAF].hlev = NumLevel;
      All[THETAF].plev = NumLevelRequest;
      All[THETAF].sfit = TRUE;
      if ( All[THETAF].hybrid == NULL )
	All[THETAF].hybrid = DoubleAlloc(Dim3GP, "All[THETAF].hybrid");
      if ( All[THETAH].hybrid == NULL )
	All[THETAH].hybrid = DoubleAlloc(Dim3GP, "All[THETAH].hybrid");
      theta(All[THETAF].hybrid, All[THETAH].hybrid, All[HALF_PRESS].hybrid, All[PS_PROG].hybrid,
	    All[TEMPERATURE].hybrid, All[TS].hybrid, NumLevel, DimGP, Dim3GP);
    }

  if ( All[GEOPOTHEIGHT].comp )
    {
      All[GEOPOTHEIGHT].hlev = NumLevel+1;
      All[GEOPOTHEIGHT].plev = NumLevelRequest;
      All[GEOPOTHEIGHT].sfit = TRUE;
      All[GEOPOTHEIGHT].hybrid = DoubleAlloc(Dim3GP+DimGP, "All[GEOPOTHEIGHT].hybrid");

      RealCopy(All[GEOPOTHEIGHT].hybrid+Dim3GP, Orography, DimGP);
      MakeGeopotHeight(All[GEOPOTHEIGHT].hybrid, All[TEMPERATURE].hybrid,
		       All[HUMIDITY].hybrid, All[HALF_PRESS].hybrid, DimGP, NumLevel);

      All[HUMIDITY].needed = All[HUMIDITY].selected;
    }

  if ( All[DPSDX].needed || All[DPSDY].needed )
    for (l = 0; l < DimGP; l++)
      {
	All[DPSDX].hybrid[l] *= All[PS_PROG].hybrid[l];
	All[DPSDY].hybrid[l] *= All[PS_PROG].hybrid[l];
      }

  if ( All[OMEGA].comp )
    {
      All[OMEGA].hlev = NumLevel+1;
      All[OMEGA].plev = NumLevelRequest;
      All[OMEGA].sfit = TRUE;
      All[OMEGA].hybrid = DoubleAlloc(Dim3GP+DimGP, "OMEGA.hybrid");

      Omega(All[OMEGA].hybrid, All[DIVERGENCE].hybrid, All[U_WIND].hybrid, All[V_WIND].hybrid,
	    All[HALF_PRESS].hybrid, All[FULL_PRESS].hybrid, All[DPSDX].hybrid, All[DPSDY].hybrid,
	    vct, DimGP, Dim3GP, NumLevel);
      
      All[DPSDX].needed = All[DPSDX].selected;
      All[DPSDY].needed = All[DPSDY].selected;
    }

  if ( All[WINDSPEED].comp )
    {
      All[WINDSPEED].hlev = NumLevel;
      All[WINDSPEED].plev = NumLevelRequest;
      All[WINDSPEED].sfit = TRUE;
      All[WINDSPEED].hybrid = DoubleAlloc(Dim3GP, "All[WINDSPEED].hybrid");

      windSpeed(All[WINDSPEED].hybrid, All[U_WIND].hybrid, All[V_WIND].hybrid, Dim3GP);
    }

  if ( All[RHUMIDITY].comp )
    {
      All[RHUMIDITY].hlev = NumLevel;
      All[RHUMIDITY].plev = NumLevelRequest;
      All[RHUMIDITY].sfit = FALSE;
      All[RHUMIDITY].hybrid = DoubleAlloc(Dim3GP, "All[RHUMIDITY].hybrid");

      sh2rh(All[HUMIDITY].hybrid, All[RHUMIDITY].hybrid, All[TEMPERATURE].hybrid, NumLevel,
	    DimGPOut, LevelRequest, All[FULL_PRESS].hybrid);

      All[TEMPERATURE].needed = All[TEMPERATURE].selected;
      All[HUMIDITY].needed    = All[HUMIDITY].selected;
    }

  if ( All[PS].comp )
    {
      All[PS].hlev = 1;
      All[PS].plev = 1;
      All[PS].sfit = TRUE; /* ??? */
      All[PS].hybrid = DoubleAlloc(DimGP, "All[PS].hybrid");
      for (l = 0; l < DimGP; l++) All[PS].hybrid[l] = exp(All[LNPS].hybrid[l]);
    }

  if ( All[SLP].comp )
    {
      All[SLP].hlev = 1;
      All[SLP].plev = 1;
      All[SLP].sfit = TRUE;
      All[SLP].hybrid = DoubleAlloc(DimGP, "All[SLP].hybrid");

      extra_P(All[SLP].hybrid, All[HALF_PRESS].hybrid + Dim3GP,
	      All[FULL_PRESS].hybrid + Dim3GP - DimGP , Orography,
	      All[TEMPERATURE].hybrid + Dim3GP - DimGP , DimGP);
      All[TEMPERATURE].needed = All[TEMPERATURE].selected || All[GEOPOTHEIGHT].selected;
    }

  if ( All[PRECIP].comp )
    {
      All[PRECIP].hlev = All[PRECIP].plev = 1;
      All[PRECIP].sfit = FALSE;
      All[PRECIP].hybrid = DoubleAlloc(DimGP, "PRECIP.hybrid");
      Add2Vectors(All[PRECIP].hybrid, All[142].hybrid, All[143].hybrid, DimGP);
    }

  if ( All[NET_TOP].comp )
    {
      All[NET_TOP].hlev = All[NET_TOP].plev = 1;
      All[NET_TOP].sfit = FALSE;
      All[NET_TOP].hybrid = DoubleAlloc(DimGP, "NET_TOP.hybrid");
      Add2Vectors(All[NET_TOP].hybrid, All[178].hybrid, All[179].hybrid, DimGP);
    }

  if ( All[NET_BOT].comp )
    {
      All[NET_BOT].hlev = All[NET_BOT].plev = 1;
      All[NET_BOT].sfit = FALSE;
      All[NET_BOT].hybrid = DoubleAlloc(DimGP, "NET_BOT.hybrid");
      Add2Vectors(All[NET_BOT].hybrid, All[176].hybrid, All[177].hybrid, DimGP);
    }

  if ( All[NET_HEAT].comp )
    {
      All[NET_HEAT].hlev = All[NET_HEAT].plev = 1;
      All[NET_HEAT].sfit = FALSE;
      All[NET_HEAT].hybrid = DoubleAlloc(DimGP, "NET_HEAT.hybrid");
      /*
      if ( Source == S_ECHAM5 )
	{
	  MultVectorScalar(All[NET_HEAT].hybrid, All[218].hybrid, (-3.345e5), DimGP,
	                   All[218].nmiss, All[218].missval);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[176].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[177].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[146].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[147].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[206].hybrid, DimGP);
	  Sub2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[208].hybrid, DimGP);
	  Sub2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[209].hybrid, DimGP);
	}
      else
      */
	{
	  MultVectorScalar(All[NET_HEAT].hybrid, All[218].hybrid, C_TIMES_RHOH2O, DimGP,
			   All[218].nmiss, All[218].missval);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[176].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[177].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[146].hybrid, DimGP);
	  Add2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[147].hybrid, DimGP);
	  Sub2Vectors(All[NET_HEAT].hybrid, All[NET_HEAT].hybrid, All[220].hybrid, DimGP);
	}
    }

  if ( All[NET_WATER].comp )
    {
      All[NET_WATER].hlev = All[NET_WATER].plev = 1;
      All[NET_WATER].sfit = FALSE;
      All[NET_WATER].hybrid = DoubleAlloc(DimGP, "NET_WATER.hybrid");
      Sub2Vectors(All[NET_WATER].hybrid, All[182].hybrid ,      All[160].hybrid, DimGP);
      Add2Vectors(All[NET_WATER].hybrid, All[NET_WATER].hybrid, All[142].hybrid, DimGP);
      Add2Vectors(All[NET_WATER].hybrid, All[NET_WATER].hybrid, All[143].hybrid, DimGP);
    }

  if ( All[LOW_WATER].comp)
    {
      All[LOW_WATER].hlev = All[LOW_WATER].plev = 1;
      All[LOW_WATER].sfit = FALSE;
      All[LOW_WATER].hybrid = DoubleAlloc(DimGP, "All[LOW_WATER].hybrid");
      LayerWater(All[222].hybrid, All[LOW_WATER].hybrid, 75000.,101300., DimGP, HalfLevels, vct);
    }

  if ( All[MID_WATER].comp)
    {
      All[MID_WATER].hlev = All[MID_WATER].plev = 1;
      All[MID_WATER].sfit = FALSE;
      All[MID_WATER].hybrid = DoubleAlloc(DimGP, "All[MID_WATER].hybrid");
      LayerWater(All[222].hybrid, All[MID_WATER].hybrid, 46000., 73000., DimGP, HalfLevels, vct);
    }

  if ( All[HIH_WATER].comp )
    {
      All[HIH_WATER].hlev = All[HIH_WATER].plev = 1;
      All[HIH_WATER].sfit = FALSE;
      All[HIH_WATER].hybrid = DoubleAlloc(DimGP, "All[HIH_WATER].hybrid");
      LayerWater(All[222].hybrid, All[HIH_WATER].hybrid,  5000., 44000., DimGP, HalfLevels, vct);
    }

  if ( All[ALL_WATER].comp )
    {
      All[ALL_WATER].hlev = All[ALL_WATER].plev = 1;
      All[ALL_WATER].sfit = FALSE;
      All[ALL_WATER].hybrid = DoubleAlloc(DimGP, "All[ALL_WATER].hybrid");
      LayerWater(All[222].hybrid, All[ALL_WATER].hybrid,  5000.,101300., DimGP, HalfLevels, vct);
    }

  if ( All[LOW_CLOUD].comp )
    {
      All[LOW_CLOUD].hlev = All[LOW_CLOUD].plev = 1;
      All[LOW_CLOUD].sfit = FALSE;
      All[LOW_CLOUD].hybrid = DoubleAlloc(DimGP, "All[LOW_CLOUD].hybrid");
      LayerCloud(All[223].hybrid, All[LOW_CLOUD].hybrid, 75000.,101300., DimGP, HalfLevels, vct);
    }

  if ( All[MID_CLOUD].comp )
    {
      All[MID_CLOUD].hlev = All[MID_CLOUD].plev = 1;
      All[MID_CLOUD].sfit = FALSE;
      All[MID_CLOUD].hybrid = DoubleAlloc(DimGP, "All[MID_CLOUD].hybrid");
      LayerCloud(All[223].hybrid, All[MID_CLOUD].hybrid, 46000., 73000., DimGP, HalfLevels, vct);
    }

  if ( All[HIH_CLOUD].comp )
    {
      All[HIH_CLOUD].hlev = All[HIH_CLOUD].plev = 1;
      All[HIH_CLOUD].sfit = FALSE;
      All[HIH_CLOUD].hybrid = DoubleAlloc(DimGP, "All[HIH_CLOUD].hybrid");
      LayerCloud(All[223].hybrid, All[HIH_CLOUD].hybrid,  5000., 44000., DimGP, HalfLevels, vct);
    }

  if ( All[SW_CLF].comp )
    {
      All[SW_CLF].hlev = All[SW_CLF].plev = 1;
      All[SW_CLF].sfit = FALSE;
      All[SW_CLF].hybrid = DoubleAlloc(DimGP, "SW_CLF.hybrid");
      Sub2Vectors(All[SW_CLF].hybrid, All[178].hybrid, All[224].hybrid, DimGP);
    }

  if ( All[SW_BOT_CLF].comp )
    {
      All[SW_BOT_CLF].hlev = All[SW_BOT_CLF].plev = 1;
      All[SW_BOT_CLF].sfit = FALSE;
      All[SW_BOT_CLF].hybrid = DoubleAlloc(DimGP, "All[SW_BOT_CLF].hybrid");
      Sub2Vectors(All[SW_BOT_CLF].hybrid, All[176].hybrid, All[185].hybrid, DimGP);
    }

  if ( All[SW_TOP_CLF].comp )
    {
      All[SW_TOP_CLF].hlev = All[SW_TOP_CLF].plev = 1;
      All[SW_TOP_CLF].sfit = FALSE;
      All[SW_TOP_CLF].hybrid = DoubleAlloc(DimGP, "All[SW_TOP_CLF].hybrid");
      Sub2Vectors(All[SW_TOP_CLF].hybrid, All[178].hybrid, All[187].hybrid, DimGP);
    }

  if ( All[LW_CLF].comp )
    {
      All[LW_CLF].hlev = All[LW_CLF].plev = 1;
      All[LW_CLF].sfit = FALSE;
      All[LW_CLF].hybrid = DoubleAlloc(DimGP, "LW_CLF.hybrid");
      Sub2Vectors(All[LW_CLF].hybrid, All[179].hybrid, All[225].hybrid, DimGP);
    }

  if ( All[LW_BOT_CLF].comp )
    {
      All[LW_BOT_CLF].hlev = All[LW_BOT_CLF].plev = 1;
      All[LW_BOT_CLF].sfit = FALSE;
      All[LW_BOT_CLF].hybrid = DoubleAlloc(DimGP, "All[LW_BOT_CLF].hybrid");
      Sub2Vectors(All[LW_BOT_CLF].hybrid, All[177].hybrid, All[186].hybrid, DimGP);
    }

  if ( All[LW_TOP_CLF].comp )
    {
      All[LW_TOP_CLF].hlev = All[LW_TOP_CLF].plev = 1;
      All[LW_TOP_CLF].sfit = FALSE;
      All[LW_TOP_CLF].hybrid = DoubleAlloc(DimGP, "All[LW_TOP_CLF].hybrid");
      Sub2Vectors(All[LW_TOP_CLF].hybrid, All[179].hybrid, All[188].hybrid, DimGP);
    }

  if ( All[NET_CLF].comp )
    {
      All[NET_CLF].hlev = All[NET_CLF].plev = 1;
      All[NET_CLF].sfit = FALSE;
      All[NET_CLF].hybrid = DoubleAlloc(DimGP, "NET_CLF.hybrid");
      Add2Vectors(All[NET_CLF].hybrid, All[178].hybrid, All[179].hybrid, DimGP);
      Sub2Vectors(All[NET_CLF].hybrid, All[NET_CLF].hybrid, All[224].hybrid, DimGP);
      Sub2Vectors(All[NET_CLF].hybrid, All[NET_CLF].hybrid, All[225].hybrid, DimGP);
    }

  if ( All[SW_ATM].comp )
    {
      All[SW_ATM].hlev = All[SW_ATM].plev = 1;
      All[SW_ATM].sfit = FALSE;
      All[SW_ATM].hybrid = DoubleAlloc(DimGP, "All[SW_ATM].hybrid");
      Sub2Vectors(All[SW_ATM].hybrid, All[178].hybrid, All[176].hybrid,DimGP);
    }

  if ( All[LW_ATM].comp )
    {
      All[LW_ATM].hlev = All[LW_ATM].plev = 1;
      All[LW_ATM].sfit = FALSE;
      All[LW_ATM].hybrid = DoubleAlloc(DimGP, "All[LW_ATM].hybrid");
      Sub2Vectors(All[LW_ATM].hybrid, All[179].hybrid, All[177].hybrid,DimGP);
    }

  if ( All[NET_ATM].comp )
    {
      All[NET_ATM].hlev = All[NET_ATM].plev = 1;
      All[NET_ATM].sfit = FALSE;
      All[NET_ATM].hybrid = DoubleAlloc(DimGP, "All[NET_ATM].hybrid");
      Add2Vectors(All[NET_ATM].hybrid, All[178].hybrid, All[179].hybrid,DimGP);
      Sub2Vectors(All[NET_ATM].hybrid, All[NET_ATM].hybrid, All[176].hybrid,DimGP);
      Sub2Vectors(All[NET_ATM].hybrid, All[NET_ATM].hybrid, All[177].hybrid,DimGP);
    }

  if ( All[SURF_RUNOFF].comp )
    {
      All[SURF_RUNOFF].hlev = All[SURF_RUNOFF].plev = 1;
      All[SURF_RUNOFF].sfit = FALSE;
      All[SURF_RUNOFF].hybrid = DoubleAlloc(DimGP, "All[SURF_RUNOFF].hybrid");
      Sub2Vectors(All[SURF_RUNOFF].hybrid, All[182].hybrid, All[221].hybrid,DimGP);
      Add2Vectors(All[SURF_RUNOFF].hybrid, All[SURF_RUNOFF].hybrid, All[142].hybrid,DimGP);
      Add2Vectors(All[SURF_RUNOFF].hybrid, All[SURF_RUNOFF].hybrid, All[143].hybrid,DimGP);
    }

  if ( All[FRESH_WATER].comp )
    {
      All[FRESH_WATER].hlev = All[FRESH_WATER].plev = 1;
      All[FRESH_WATER].sfit = FALSE;
      All[FRESH_WATER].hybrid = DoubleAlloc(DimGP, "All[FRESH_WATER].hybrid");
      Add2Vectors(All[FRESH_WATER].hybrid, All[142].hybrid, All[143].hybrid, DimGP);
      Add2Vectors(All[FRESH_WATER].hybrid, All[FRESH_WATER].hybrid, All[182].hybrid, DimGP);
    }
}

/*
 *  Process Model Level data
 */
void processML(void)
{
  static char func[] = "processML";
  int code,l,i;
  long fieldSize = 0;
  double *var_spectral, *div_spectral, *vor_spectral;
  int lindex, ilevel, nlevel;
  int offset;
  int leveltype;
  int nmiss;
  double *pressureLevel = NULL;

  MeanCount++;
  TermCount++;

  CheckContent(All, TermCount);

  if ( MeanCount == 1 )
    {
      if ( Debug ) Message(func, "TermCount = %d MeanCount = %d", TermCount, MeanCount);
      StartDate = OldDate;
    }

  if ( TermCount  > 120 ) Debug = 0;

  /* ============================== */
  /* Computations in spectral space */
  /* ============================== */

  if ( All[U_WIND].comp || All[V_WIND].comp )
    {
      /*
	double tmp[1000000];
	*/
      All[U_WIND].hlev = All[V_WIND].hlev = All[DIVERGENCE].hlev;
      All[U_WIND].plev = All[V_WIND].plev = All[DIVERGENCE].plev;
      All[U_WIND].sfit = All[V_WIND].sfit = TRUE;
      All[U_WIND].spectral = DoubleAlloc(Dim3SP, "All[U_WIND].spectral");
      All[V_WIND].spectral = DoubleAlloc(Dim3SP, "All[V_WIND].spectral");

      if ( All[DIVERGENCE].spectral == NULL ) gp2sp(DIVERGENCE);
      if ( All[VORTICITY].spectral == NULL )  gp2sp(VORTICITY);
      /*
	 sp2sp(All[DIVERGENCE].spectral, Truncation, tmp, Truncation-1);
	 sp2sp(tmp, Truncation-1, All[DIVERGENCE].spectral, Truncation);
	 sp2sp(All[VORTICITY].spectral, Truncation, tmp, Truncation-1);
	 sp2sp(tmp, Truncation-1, All[VORTICITY].spectral, Truncation);
	 printf("dv2uv: \n");
	 */	 
      dv2uv(All[DIVERGENCE].spectral, All[VORTICITY].spectral,
	    All[U_WIND].spectral, All[V_WIND].spectral,
	    dv2uv_f1, dv2uv_f2,
	    Truncation, DimSP, All[DIVERGENCE].hlev);
      /*
	 printf("dv2uv:  DimSP = %d, DimFC = %d\n", DimSP, DimFC);
	 {
	   int i, m, n;
           i=0;
	   for ( m = 0; m <= Truncation; m++ )
	     for ( n = m; n <= Truncation; n++ )
	       {
		 printf("m= %d n = %d, i = %d val = %g %g %g %g\n",
			m, n, i, All[DIVERGENCE].spectral[i], All[VORTICITY].spectral[i],
			All[U_WIND].spectral[i],    All[V_WIND].spectral[i]);
		 printf("m= %d n = %d, i = %d val = %g %g %g %g\n",
			m, n, i+1, All[DIVERGENCE].spectral[i+1], All[VORTICITY].spectral[i+1],
			All[U_WIND].spectral[i+1],    All[V_WIND].spectral[i+1]);
		 i += 2;
	       }
	 }
	 */
   }

   if ( All[VELOPOT].comp && Type < 30 )
     {
       All[VELOPOT].hlev = All[DIVERGENCE].hlev;
       All[VELOPOT].plev = All[DIVERGENCE].plev;
       All[VELOPOT].spectral = DoubleAlloc(Dim3SP, "All[VELOPOT].spectral");

       if ( All[DIVERGENCE].spectral == NULL ) gp2sp(DIVERGENCE);

       dv2ps(All[DIVERGENCE].spectral, All[VELOPOT].spectral,
	     All[DIVERGENCE].hlev, Truncation);
     }

   if ( All[STREAM].comp && Type < 30 )
     {
       All[STREAM].hlev = All[VORTICITY].hlev;
       All[STREAM].plev = All[VORTICITY].plev;
       All[STREAM].spectral = DoubleAlloc(Dim3SP, "All[STREAM].spectral");

       if ( All[VORTICITY].spectral == NULL ) gp2sp(VORTICITY);

       dv2ps(All[VORTICITY].spectral, All[STREAM].spectral,
	     All[VORTICITY].hlev, Truncation);
     }

   if ( All[VORTICITY].spectral && !All[VORTICITY].needed )
        All[VORTICITY].spectral = FreeMemory(All[VORTICITY].spectral);

   if ( All[DIVERGENCE].spectral && !All[DIVERGENCE].needed )
        All[DIVERGENCE].spectral = FreeMemory(All[DIVERGENCE].spectral);

  /* --------------------------- */
  /*  Output of spectral fields  */
  /* --------------------------- */

  if ( Type == 0 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].spectral )
	      {
		leveltype = zaxisInqType(All[code].ozaxisID);
		if ( (leveltype == ZAXIS_HYBRID) && (All[code].ozaxisID != All[code].izaxisID) )
		  {
		    nlevel = zaxisInqSize(All[code].ozaxisID);
		    for ( lindex = 0; lindex < nlevel; lindex++ )
		      {
			ilevel = (int) zaxisInqLevel(All[code].ozaxisID, lindex);
			offset = (ilevel-1)*DimSP;
			streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].spectral+offset, 0);
		      }
		  }
		else
		  streamWriteVar(ostreamID,  All[code].ovarID,  All[code].spectral, 0);
	      }
	    else
	      Error(func, "Code %d not available on spectral space!", code);
	  }

      FreeSpectral(All);
      return;
    }

  /* ------------------------------- */
  /*  Computations in fourier space  */
  /* ------------------------------- */

  if ( Type >= 10 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].spectral )
	  {
	    if ( All[code].fourier == NULL )
	      {
		fieldSize = All[code].hlev * DimFC;
		All[code].fourier = DoubleAlloc(fieldSize,
						FieldName(code,"fourier"));
		sp2fc(All[code].spectral, All[code].fourier, poli,
		      All[code].hlev, Latitudes, Fouriers, Truncation);
	      }
	    if ( code != LNPS )
	      All[code].spectral = FreeMemory(All[code].spectral);
	  }

      /*    if (Type < 60) poli = FreeMemory(poli); */
      /*    if (Type < 50) pol2 = FreeMemory(pol2); */
      /*    if (Type < 50) pol3 = FreeMemory(pol3); */

      if ( All[U_WIND].needed && All[U_WIND].fourier )
	scaluv(All[U_WIND].fourier, rcoslat, Latitudes, Fouriers*NumLevel);
      if ( All[V_WIND].needed && All[V_WIND].fourier )
	scaluv(All[V_WIND].fourier, rcoslat, Latitudes, Fouriers*NumLevel);

      if ( All[DPSDX].needed )
	{
	  All[DPSDX].hlev = 1;
	  All[DPSDX].plev = 1;
	  All[DPSDX].sfit = FALSE;
	  All[DPSDX].fourier = DoubleAlloc(DimFC, "All[DPSDX].fourier");
	  if ( All[LNPS].fourier == NULL )  gp2sp(LNPS);
	  Derivate(All[LNPS].fourier, All[DPSDX].fourier, 1, Waves, Latitudes, DerivationFactor);
	}
      if ( All[DPSDY].needed )
	{
	  All[DPSDY].hlev = 1;
	  All[DPSDY].plev = 1;
	  All[DPSDY].sfit = FALSE;
	  All[DPSDY].fourier = DoubleAlloc(DimFC, "All[DPSDY].fourier");
	  if ( All[LNPS].spectral == NULL )  gp2sp(LNPS);
	  sp2fc(All[LNPS].spectral, All[DPSDY].fourier, pdev,
		All[DPSDY].hlev, Latitudes, Fouriers, Truncation);
	}
    }

  FreeSpectral(All);

  /* -------------------------- */
  /*  Output of fourier fields  */
  /* -------------------------- */

  if ( Type == 10 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].fourier, 0);
	    else
	      Error(func, "Code %d not available on fourier space!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ----------------------- */
  /*  Output of zonal means  */
  /* ----------------------- */

  if ( Type == 11 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      {
		leveltype = zaxisInqType(All[code].ozaxisID);
		if ( (leveltype == ZAXIS_HYBRID) && (All[code].ozaxisID != All[code].izaxisID) )
		  {
		    nlevel = zaxisInqSize(All[code].ozaxisID);
		    for ( lindex = 0; lindex < nlevel; lindex++ )
		      {
			ilevel = (int) zaxisInqLevel(All[code].ozaxisID, lindex);
			offset = (ilevel-1)*DimFC;
			streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].fourier+offset, 0);
		      }
		  }
		else
		  {
		    nlevel = zaxisInqSize(All[code].ozaxisID);
		    for ( lindex = 0; lindex < nlevel; lindex++ )
		      {
			offset = lindex*DimFC;
			streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].fourier+offset, 0);
		      }
		  }
	      }
	    else
	      Error(func, "Code %d not available on zonal mean!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ------------------------------ */
  /*  Transformation to gridpoints  */
  /* ------------------------------ */

  if ( Type >= 20 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].fourier )
	  {
	    if ( All[code].hybrid == NULL )
	      {
		fieldSize = DimGP * All[code].hlev;
		All[code].hybrid = DoubleAlloc(fieldSize,
					       FieldName(code,"hybrid"));
		FC2GP(All[code].fourier,All[code].hybrid,
		      Latitudes,Longitudes,All[code].hlev,Fouriers);
	      }
	    All[code].fourier = FreeMemory(All[code].fourier);
	  }

      if ( All[PS_PROG].comp && All[PS_PROG].hybrid == NULL )
	{
	  All[PS_PROG].hybrid = DoubleAlloc(DimGP, "PS_PROG");
	  if ( All[LNPS].hybrid )
	    {
	      for (l = 0; l < DimGP; l++) All[PS_PROG].hybrid[l] = exp(All[LNPS].hybrid[l]);
	    }
	  else if ( All[PS].hybrid )
	    {
	      Warning(func, "log surface pressure (code 152) not found - using surface pressure (code 134)!");
	      RealCopy(All[PS_PROG].hybrid, All[PS].hybrid, DimGP);
	    }
	  else
	    {
	      Error(func, "surface pressure not found!");
	    }
	}
      All[LNPS].needed = All[LNPS].selected;
      
      if ( Orography == NULL )
	{
	  Orography = DoubleAlloc(DimGP , "Orography");
	  if ( All[GEOPOTENTIAL].hybrid )
	    RealCopy(Orography, All[GEOPOTENTIAL].hybrid, DimGP);
	  else
	    {
	      if ( All[GEOPOTENTIAL].selected || Type >= 30 )
		{
		  Warning(func, "Orography not found - using zero orography!");
		  RealZero(Orography, DimGP);
		}
	    }
	}
      All[GEOPOTENTIAL].needed = All[GEOPOTENTIAL].selected;

      EchamCompGP();
    }

  FreeFourier(All);

  if ( Type == 20 )
    {
      /* ----------------------- */
      /*  Means on hybrid grids  */
      /* ----------------------- */

      if ( Mean )
	{
	  for ( code = 0; code < MaxCodes; code++ )
	    {
	      if ( All[code].selected && All[code].hybrid )
		{
		  fieldSize = DimGP * All[code].hlev;

		  if ( All[code].mean == NULL )
		    All[code].mean = DoubleAlloc(fieldSize, FieldName(code, "mean"));

		  if ( Mean > 1 && All[code].variance == NULL )
		    All[code].variance = DoubleAlloc(fieldSize, FieldName(code, "variance"));

		  if ( MeanCount == 1 )
		    {
		      RealCopy(All[code].mean, All[code].hybrid, fieldSize);
		      if ( Mean > 1 )
			IniQuaSum(All[code].variance, All[code].hybrid, fieldSize);
		    }
		  else
		    {
		      AddVector(All[code].mean, All[code].hybrid, fieldSize,
				&All[code].nmiss, All[code].missval);
		      if ( Mean > 1 )
			AddQuaSum(All[code].variance, All[code].hybrid, fieldSize);
		    }

		  if ( EndOfInterval )
		    {
		      if ( All[code].samp == NULL )
			MultVectorScalar(All[code].hybrid, All[code].mean, 1.0/MeanCount, fieldSize,
					 All[code].nmiss, All[code].missval);
		      else
			DivVectorIvector(All[code].hybrid, All[code].mean, All[code].samp, fieldSize,
					 &All[code].nmiss, All[code].missval);

		      if ( Mean > 1 )
			VarQuaSum(All[code].variance, All[code].hybrid, fieldSize, MeanCount);
		    }
		}
	    }
	}

      /* ---------------------------- */
      /* Output of hybrid level grids */
      /* ---------------------------- */

      if ( Mean == 0 || EndOfInterval )
	{
	  for ( code = 0; code < MaxCodes; code++ )
	    if ( All[code].selected )
	      {
		if ( All[code].hybrid == NULL )
		  Error(func, "Internal problem. Code %d not allocated!", code);

		leveltype = zaxisInqType(All[code].ozaxisID);
		if ( (leveltype == ZAXIS_HYBRID) && (All[code].ozaxisID != All[code].izaxisID) )
		  {
		    nlevel = zaxisInqSize(All[code].ozaxisID);
		    for ( lindex = 0; lindex < nlevel; lindex++ )
		      {
			ilevel = (int) zaxisInqLevel(All[code].ozaxisID, lindex);
			offset = (ilevel-1)*DimGP;
			if ( Mean != 2 )
			  streamWriteVarSlice(ostreamID,  All[code].ovarID,  lindex,
					      All[code].hybrid+offset, All[code].nmiss);
			if ( Mean >= 2 )
			  streamWriteVarSlice(ostreamID2, All[code].ovarID2, lindex,
					      All[code].variance+offset, All[code].nmiss);
		      }
		  }
		else
		  {
		    if ( Mean != 2 )
		      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].hybrid, All[code].nmiss);
		    if ( Mean >= 2 )
		      streamWriteVar(ostreamID2, All[code].ovarID2, All[code].variance, All[code].nmiss);
		  }
	      }

	  FreeSamp(All);
	}

      FreeHybrid(All);
      return;
    }

  /* -------------------------------------- */
  /* Vertical interpolation / extrapolation */
  /* -------------------------------------- */

  if ( Type >= 30 )
    {
      if ( vert_index == NULL )
	vert_index = (int *) malloc(NumLevelRequest*DimGP*sizeof(int));

      if ( unitsel )
	{
	  if ( p_of_height == NULL ) p_of_height = DoubleAlloc(NumLevelRequest,"p_of_height");
	  h2p(p_of_height, LevelRequest, NumLevelRequest);
	  pressureLevel = p_of_height;
	}
      else
	{
	  pressureLevel = LevelRequest;
	}

      genind(vert_index, pressureLevel, All[FULL_PRESS].hybrid, DimGP, NumLevelRequest, NumLevel);

      nmiss = 0;
      if ( ! Extrapolate )
	{
	  if ( pnmiss == NULL ) pnmiss = (int *) malloc(NumLevelRequest*sizeof(int));  
	  genindmiss(vert_index, pressureLevel, DimGP, NumLevelRequest, All[PS_PROG].hybrid, pnmiss);
	  for ( i = 0; i < NumLevelRequest; i++ ) nmiss += pnmiss[i];
	}

      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].hybrid )
	  {
	    leveltype = zaxisInqType(All[code].izaxisID);
	    if ( All[code].hlev == 1 || leveltype != ZAXIS_HYBRID || (All[code].hlev < NumLevel) )
	      {
		if ( All[code].grid ) FreeMemory(All[code].grid);
		All[code].grid = All[code].hybrid;
		All[code].hybrid = NULL;
	      }
	    else
	      {
		if ( All[code].grid == NULL )
		  {
		    fieldSize = DimGP * NumLevelRequest;
		    All[code].grid = DoubleAlloc(fieldSize, FieldName(code, "grid"));
		  }

		if ( code == TEMPERATURE )
		  {
		    interp_T(Orography, All[TEMPERATURE].hybrid, All[TEMPERATURE].grid,
			     All[FULL_PRESS].hybrid, All[HALF_PRESS].hybrid, vert_index,
			     pressureLevel, NumLevelRequest, DimGP, NumLevel, All[code].missval);
		  }
		else if ( code == GEOPOTHEIGHT )
		  {
		    if ( All[TEMPERATURE].hybrid == NULL ) Error(func, "Code  130 not found!");
		    interp_Z(Orography, All[GEOPOTHEIGHT].hybrid, All[GEOPOTHEIGHT].grid,
			     All[FULL_PRESS].hybrid, All[HALF_PRESS].hybrid, vert_index,
			     All[TEMPERATURE].hybrid,
			     pressureLevel, NumLevelRequest, DimGP, NumLevel, All[code].missval);
		  }
		else
		  {
		    interp_X(All[code].hybrid, All[code].grid, All[FULL_PRESS].hybrid,
			     vert_index, pressureLevel, NumLevelRequest, DimGP, NumLevel, All[code].missval);
		  }

		if ( ! Extrapolate ) All[code].nmiss = nmiss;

		if ( code != TEMPERATURE )
		  All[code].hybrid = FreeMemory(All[code].hybrid);
	      }
	  }
    }

  All[TEMPERATURE].needed = All[TEMPERATURE].selected;
  FreeHybrid(All);
  if ( All[HALF_PRESS].hybrid )
    All[HALF_PRESS].hybrid = FreeMemory(All[HALF_PRESS].hybrid);

  /* -------------------------------- */
  /*  Output of pressure level grids  */
  /* -------------------------------- */

  if ( Type == 30 && Mean == 0 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected && All[code].grid )
	  {
	    streamWriteVar(ostreamID,  All[code].ovarID,  All[code].grid, All[code].nmiss);
	  }

      FreeGrid(All);
      return;
    }

  /* ---------------------- */
  /*  Computation of Means  */
  /* ---------------------- */

  if ( Type >= 30 && Mean )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].needed && All[code].grid )
	{
	  fieldSize = DimGP * All[code].plev;

	  if ( All[code].mean == NULL )
	    All[code].mean = DoubleAlloc(fieldSize, FieldName(code, "mean"));

	  if ( MeanCount == 1 )
	    RealCopy(All[code].mean, All[code].grid, fieldSize);
	  else
	    AddVector(All[code].mean, All[code].grid, fieldSize,
		      &All[code].nmiss, All[code].missval);

	  if ( EndOfInterval )
	    {
	      if ( All[code].samp == NULL )
		MultVectorScalar(All[code].mean, All[code].mean, 1.0/MeanCount, fieldSize,
				 All[code].nmiss, All[code].missval);
	      else
		DivVectorIvector(All[code].mean, All[code].mean, All[code].samp, fieldSize,
				 &All[code].nmiss, All[code].missval);
	    }
	}

  /* -------------------------- */
  /*  Computation of Variances  */
  /* -------------------------- */

  if ( Type >= 30 && Mean > 1 )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].needed && All[code].mean )
	{
	  fieldSize = DimGP * All[code].plev;

	  if (All[code].variance == NULL)
	    All[code].variance = DoubleAlloc(fieldSize, FieldName(code, "var"));

	  if ( MeanCount == 1 )
	    IniQuaSum(All[code].variance, All[code].grid, fieldSize);
	  else
	    AddQuaSum(All[code].variance, All[code].grid, fieldSize);

	  if ( EndOfInterval )
	    VarQuaSum(All[code].variance, All[code].mean, fieldSize, MeanCount);
	}

   if ( Mean && !EndOfInterval )
     {
       FreeGrid(All);
       return;
     }

  /* --------------------------------------------- */
  /*  Output of pressure level means and variances */
  /* --------------------------------------------- */

  if ( Type == 30 && Mean )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected && All[code].mean )
	  {
	    if ( Mean != 2 )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].mean, All[code].nmiss);
	    if ( Mean >= 2 )
	      streamWriteVar(ostreamID2, All[code].ovarID2, All[code].variance, All[code].nmiss);
	  }

      FreeSamp(All);
      FreeGrid(All);
      return;
    }


  /* ------------------ */
  /*  Free mean fields  */
  /* ------------------ */

  if ( Type >= 40 && Mean )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].mean )
	{
	  if ( All[code].variance ) All[code].variance = FreeMemory(All[code].variance);
	  if ( All[code].grid )     All[code].grid     = FreeMemory(All[code].grid);
	  All[code].grid = All[code].mean;
	  All[code].mean = NULL;
	}

  /* --------------------------------- */
  /*  Transformation to fourier space  */
  /* --------------------------------- */

  if ( Type >= 40 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].grid && (All[code].sfit || Type < 70) )
	  {
	    if ( All[code].nmiss > 0 )
	      Error(func, "Missing values for code %d unsupported with TYPE > 30!", code);

	    if ( All[code].fourier == NULL )
	      {
		fieldSize = DimFC * All[code].plev;
		All[code].fourier = DoubleAlloc(fieldSize,FieldName(code,"fourier"));
	      }

	    GP2FC(All[code].grid, All[code].fourier,
		  Latitudes, Longitudes, All[code].plev, Fouriers);

	    if ( All[code].grid && (All[code].sfit || Type < 70) )
	      All[code].grid = FreeMemory(All[code].grid);
	  }
    }

  for (code = 0; code < MaxCodes; code++)
    if (All[code].grid && (All[code].sfit || Type < 70))
      All[code].grid = FreeMemory(All[code].grid);

  /* -------------------------- */
  /*  Output of fourier fields  */
  /* -------------------------- */

  if ( Type == 40 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].fourier, All[code].nmiss);
	    else
	      Error(func, "Code %d not available on fourier space!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* --------------------- */
  /* Output of zonal means */
  /* --------------------- */

  if ( Type == 41 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      {
		nlevel = zaxisInqSize(All[code].ozaxisID);
		for ( lindex = 0; lindex < nlevel; lindex++ )
		  {
		    offset = lindex*DimFC;
		    streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].fourier+offset, All[code].nmiss);
		  }
	      }
	    else
	      Error(func, "Code %d not available on zonal mean!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ---------------------------------- */
  /*  Transformation to spectral space  */
  /* ---------------------------------- */

  if ( Type >= 50 )
    {
      if ( All[U_WIND].needed && All[U_WIND].fourier )
	scaluv(All[U_WIND].fourier, coslat, Latitudes, Fouriers*NumLevelRequest);
      if ( All[V_WIND].needed && All[V_WIND].fourier )
	scaluv(All[V_WIND].fourier, coslat, Latitudes, Fouriers*NumLevelRequest);

      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].fourier )
	  {
	    if ( All[code].spectral == NULL )
	      {
		fieldSize = All[code].plev * DimSPOut;
		All[code].spectral = DoubleAlloc(fieldSize, FieldName(code,"spectral"));
	      }
	    if ( TruncationOut < Truncation )
	      {
		fieldSize = All[code].plev * DimSP;
		var_spectral = DoubleAlloc(fieldSize,"var_spectral");
		fc2sp(All[code].fourier,var_spectral,
		      pold,All[code].plev,Latitudes,Fouriers,Truncation);
		for (l = 0; l < All[code].plev; l++)
		  sp2sp(var_spectral+l*DimSP,Truncation,
				   All[code].spectral+l*DimSPOut,TruncationOut);
		var_spectral = FreeMemory(var_spectral);
	      }
	    else
	      {
		fc2sp(All[code].fourier,All[code].spectral,
		      pold,All[code].plev,Latitudes,Fouriers,Truncation);
	      }
	  }

      if ( All[DIVERGENCE].needed || All[VORTICITY].needed ||
	   All[VELOPOT].needed    || All[STREAM].needed )
	{
	  if ( All[DIVERGENCE].spectral == NULL )
	    All[DIVERGENCE].spectral = DoubleAlloc(DimSPOut*NumLevelRequest,
						   "All[DIVERGENCE].spectral");
	  if ( All[VORTICITY].spectral == NULL )
	    All[VORTICITY].spectral = DoubleAlloc(DimSPOut*NumLevelRequest,
						  "All[VORTICITY].spectral");
	  if ( TruncationOut < Truncation )
	    {
	      fieldSize = DimSP * NumLevelRequest;
	      div_spectral = DoubleAlloc(fieldSize,"div_spectral");
	      vor_spectral = DoubleAlloc(fieldSize,"vor_spectral");
	      uv2dv(All[U_WIND].fourier, All[V_WIND].fourier,
		    div_spectral, vor_spectral,
		    pol2, pol3, NumLevelRequest, Latitudes, Truncation);
	      for (l = 0; l < NumLevelRequest; l++) {
		sp2sp(div_spectral+l*DimSP, Truncation,
				 All[DIVERGENCE].spectral+l*DimSPOut, TruncationOut);
		sp2sp(vor_spectral+l*DimSP, Truncation,
				 All[VORTICITY].spectral+l*DimSPOut, TruncationOut);
	      }
	      div_spectral = FreeMemory(div_spectral);
	      vor_spectral = FreeMemory(vor_spectral);
	    }
	  else
	    {
	      if ( (All[U_WIND].fourier == 0 || All[V_WIND].fourier == 0) && NumLevelRequest )
		Error(func, "uwind or vwind missing!");
	      uv2dv(All[U_WIND].fourier, All[V_WIND].fourier,
		    All[DIVERGENCE].spectral, All[VORTICITY].spectral,
		    pol2, pol3, NumLevelRequest, Latitudes, Truncation);
	    }
	}

      if ( All[VELOPOT].needed )
	{
	  All[VELOPOT].hlev = All[DIVERGENCE].hlev;
	  All[VELOPOT].plev = All[DIVERGENCE].plev;
	  All[VELOPOT].sfit = TRUE;
	  if ( All[VELOPOT].spectral == NULL )
	    All[VELOPOT].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[VELOPOT].spectral");
	  dv2ps(All[DIVERGENCE].spectral, All[VELOPOT].spectral, NumLevelRequest, TruncationOut);
	}

      if ( All[STREAM].needed )
	{
	  All[STREAM].hlev = All[VORTICITY].hlev;
	  All[STREAM].plev = All[VORTICITY].plev;
	  All[STREAM].sfit = TRUE;
	  if ( All[STREAM].spectral == NULL )
	    All[STREAM].spectral = DoubleAlloc(DimSPOut*NumLevelRequest, "All[STREAM].spectral");
	  dv2ps(All[VORTICITY].spectral, All[STREAM].spectral, NumLevelRequest, TruncationOut);
	}
    }

  for ( code = 0; code < MaxCodes; code++ )
    if ( All[code].fourier && (All[code].sfit || Type < 61) )
      All[code].fourier = FreeMemory(All[code].fourier);

  /* --------------------------- */
  /*  Output of spectral fields  */
  /* --------------------------- */

  if ( Type == 50 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected && All[code].spectral )
	  {
	    streamWriteVar(ostreamID, All[code].ovarID, All[code].spectral, All[code].nmiss);
	  }

      FreeSpectral(All);
      return;
    }

  /* -------------------------------*/
  /*  Computations in fourier space */
  /* -------------------------------*/

  if ( Type >= 60 )
    {
      for (code = 0; code < MaxCodes; code++)
      if (All[code].needed && All[code].spectral)
	{
	  if (All[code].fourier == NULL) {
            fieldSize = All[code].plev * DimFCOut;
            All[code].fourier = DoubleAlloc(fieldSize,
                                          FieldName(code,"fourier"));
	  }
	  sp2fc(All[code].spectral,All[code].fourier,poliOut,
		All[code].plev,LatitudesOut,FouriersOut,TruncationOut);
	}
      if ( All[U_WIND].needed && All[U_WIND].fourier )
	scaluv(All[U_WIND].fourier, rcoslatOut, LatitudesOut, FouriersOut*NumLevelRequest);
      if ( All[V_WIND].needed && All[V_WIND].fourier )
	scaluv(All[V_WIND].fourier, rcoslatOut, LatitudesOut, FouriersOut*NumLevelRequest);
    }

  FreeSpectral(All);

  /* -------------------------- */
  /*  Output of fourier fields  */
  /* -------------------------- */

  if ( Type == 60 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      streamWriteVar(ostreamID,  All[code].ovarID,  All[code].fourier, 0);
	    else
	      Error(func, "Code %d not available on fourier space!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ----------------------- */
  /*  Output of zonal means  */
  /* ----------------------- */

  if ( Type == 61 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    if ( All[code].fourier )
	      {
		nlevel = zaxisInqSize(All[code].ozaxisID);
		for ( lindex = 0; lindex < nlevel; lindex++ )
		  {
		    offset = lindex*DimFCOut;
		    streamWriteVarSlice(ostreamID, All[code].ovarID, lindex, All[code].fourier+offset, All[code].nmiss);
		  }
	      }
	    else
	      Error(func, "Code %d not available on zonal mean!", code);
	  }

      FreeFourier(All);
      return;
    }

  /* ------------------------------ */
  /*  Transformation to gridpoints  */
  /* ------------------------------ */

  if ( Type >= 70 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].needed && All[code].fourier )
	  {
	    fieldSize = All[code].plev * DimGPOut;
	    if ( All[code].grid == NULL )
	      All[code].grid = DoubleAlloc(fieldSize, FieldName(code, "grid"));

	    FC2GP(All[code].fourier, All[code].grid,
		  LatitudesOut, LongitudesOut, All[code].plev, FouriersOut);
	  }
    }

  FreeFourier(All);

  /* -------------------------------- */
  /*  Output of pressure level grids  */
  /* -------------------------------- */

  if ( Type == 70 )
    {
      for ( code = 0; code < MaxCodes; code++ )
	if ( All[code].selected )
	  {
	    streamWriteVar(ostreamID, All[code].ovarID, All[code].grid, All[code].nmiss);
	  }

      FreeSamp(All);
      FreeGrid(All);
      return;
    }
}


void PostProcess(void)
{
  if ( EndOfInterval )
    {
      if      ( OutputInterval == DAILY_INTERVAL )
	fprintf(stdout, " Processed Day %2d  Month %2d  Year %04d",
		OldDate.dy, OldDate.mo, OldDate.yr);
      else if ( OutputInterval == MONTHLY_INTERVAL )
	fprintf(stdout, " Processed Month %2d  Year %04d", OldDate.mo, OldDate.yr);
      else if ( OutputInterval == UNLIM_INTERVAL )
	fprintf(stdout, " Processed range from %6.4d-%2.2d-%2.2d to %6.4d-%2.2d-%2.2d",
		StartDate.yr, StartDate.mo, OldDate.mo,
		OldDate.yr, OldDate.mo, OldDate.dy);

      if ( Mean ) fprintf(stdout, "  (Mean of %3d Terms)\n", MeanCount);
      else        fprintf(stdout, "   Terms %3d\n", MeanCount);

      EndOfInterval = FALSE;
      MeanCount = 0;
    }
}


/* ================= */
/* switch input file */
/* ================= */
void SwitchFile()
{
  int echam4 = FALSE;
  int i, n;
  char y3, y2, y1, y0;
  char         m1, m0;
  char         d1, d0;

  streamClose(istreamID);

  if ( Multi > 0 )
    {
      i = strlen(ifile);
      if ( i < 10 )
	{
	  fprintf(stderr, " Not a valid filename: %s \n", ifile);
	  exit(1);
	}

      if ( ifile[i-3] == '.' )
	{
	  echam4 = TRUE;
	  y3 = ifile[i-9]; y2 = ifile[i-8];
	  y1 = ifile[i-7]; y0 = ifile[i-6];
	  m1 = ifile[i-5]; m0 = ifile[i-4];
	  d1 = ifile[i-2]; d0 = ifile[i-1];
	}
      else
	{
	  y3 = ifile[i-6]; y2 = ifile[i-5];
	  y1 = ifile[i-4]; y0 = ifile[i-3];
	  m1 = ifile[i-2]; m0 = ifile[i-1];
	  d1 = '0';        d0 = '1'   ;
	}

      for ( n = 0; n < DayIn; n++ )
	{
	  if ( d0 =='9' ) { d0 = '0'; d1++; }
	  else d0++;
	  if ( d1 == '3' && d0 > '0' )
	    {
	      d1 = '0'; d0 = '1';
	      if ( m1 == '0' )
		{
		  if ( m0 == '9' ) { m0 = '0'; m1 = '1'; }
		  else m0++;
		}
	      else
		{
		  if ( m0 < '2' ) m0++;
		  else
		    {
		      m1 = '0';  m0 = '1';  y0++;
		      if ( y0 > '9' ) { y0 = '0'; y1++; }
		      if ( y1 > '9' ) {
			y1 = (char) '0';
			if ( isdigit((int)y2) ) y2++;
			else                    y2 = '1';
			if ( y2 > '9' )
			  {
			    y2 = (char) '0';
			    if ( isdigit((int)y3) ) y3++;
			    else                    y3 = '1';
			  }
		      }
		    }
		}
	    }
	}

      if ( echam4 )
	{
	  ifile[i-9] = y3; ifile[i-8] = y2;
	  ifile[i-7] = y1; ifile[i-6] = y0;
	  ifile[i-5] = m1; ifile[i-4] = m0;
	  ifile[i-2] = d1; ifile[i-1] = d0;
	}
      else
	{
	  ifile[i-6] = y3; ifile[i-5] = y2;
	  ifile[i-4] = y1; ifile[i-3] = y0;
	  ifile[i-2] = m1; ifile[i-1] = m0;
	}

      Multi--;
    }

  if ( Nfiles > 0 ) ifile = ifiles[--Nfiles];

  fprintf(stderr, " Continuation file: %s\n", ifile);

  istreamID = streamOpenRead(ifile);
  if ( istreamID < 0 ) cdiError(istreamID, "Open failed on %s", ifile);

  ivlistID = streamInqVlist(istreamID);
  taxisID  = vlistInqTaxis(ivlistID);
}

int
getDate(struct Date datetime)
{
  int yr, mo, dy;
  int date;

  yr = datetime.yr;
  mo = datetime.mo;
  dy = datetime.dy;

  date = cdiEncodeDate(yr, mo, dy);

  return (date);
}

int
getTime(struct Date datetime)
{
  int hr, mn;
  int time;

  hr = datetime.hr;
  mn = datetime.mn;

  time = cdiEncodeTime(hr, mn, 0);

  return (time);
}

void
setDateTime(struct Date *datetime, int date, int time)
{
  int yr, mo, dy, hr, mn, sec;

  cdiDecodeDate(date, &yr, &mo, &dy);
  cdiDecodeTime(time, &hr, &mn, &sec);
  /*
  printf(" %d %d  %6.4d-%2.2d-%2.2d  %2.2d:%2.2d\n", date, time, yr, mo, dy, hr, mn);
  */
  datetime->yr = yr;
  datetime->mo = mo;
  datetime->dy = dy;
  datetime->hr = hr;
  datetime->mn = mn;
}

void AnalysisAddRecord(int code, int gridID, int zaxisID, int levelID, int nmiss)
{
  static char func[] = "AnalysisAddRecord";
  long fieldSize;
  int truncation;
  int dataSize;
  int dataOffset;
  int nlevel;
  int gridSize;
  int gridtype, leveltype;

  gridtype   = gridInqType(gridID);
  leveltype  = zaxisInqType(zaxisID);
  nlevel     = zaxisInqSize(zaxisID);
  gridSize   = gridInqSize(gridID);
  dataSize   = gridSize*nlevel;
  dataOffset = gridSize*levelID;

  All[code].nmiss0 += nmiss;

  if ( gridtype == GRID_SPECTRAL )
    {
      All[code].sfit = TRUE;
      All[code].hlev = NumLevelRequest;
      All[code].plev = NumLevelRequest;
      if ( nlevel > 1 && leveltype == ZAXIS_PRESSURE )
	{
	  if ( code != U_WIND && code != V_WIND )
	    {
	      fieldSize = Dim3SPOut;
	      if (All[code].spectral0 == NULL)
		All[code].spectral0 = DoubleAlloc(fieldSize, FieldName(code,"spectral"));
	      truncation = gridInqTrunc(gridID);
	      sp2sp(Field, truncation, All[code].spectral0+levelID*DimSPOut, TruncationOut);
	    }
	  else
	    {
	      fieldSize = Dim3SP;
	      if (All[code].spectral0 == NULL)
		All[code].spectral0 = DoubleAlloc(fieldSize, FieldName(code,"spectral"));
	      RealCopy(All[code].spectral0+levelID*DimSP,Field,DimSP);
	    }
	}
      else
	Error(func, "leveltype %s unsupported for spectral data", zaxisNamePtr(leveltype));
    }
  else
    {
      if ( nlevel > 1 && leveltype == ZAXIS_PRESSURE )
	{
	  All[code].sfit = TRUE;
	  fieldSize = Dim3GPOut;
	  All[code].hlev = NumLevelRequest;
	  All[code].plev = NumLevelRequest;
	  if ( All[code].grid0 == NULL )
	    All[code].grid0 = DoubleAlloc(fieldSize, FieldName(code,"grid0"));
	  RealCopy(All[code].grid0+levelID*DimGP, Field, DimGP);
	}
      else
	{
	  All[code].sfit = FALSE;
	  fieldSize = DimGPOut;
	  All[code].hlev = 1;
	  All[code].plev = 1;
	  if ( All[code].grid0 == NULL )
	    All[code].grid0 = DoubleAlloc(fieldSize, FieldName(code,"grid0"));
	  RealCopy(All[code].grid0, Field, DimGP);
	}

      if ( Mean > 0 && (nmiss > 0 || All[code].samp) )
	{
	  int i;
	  if ( All[code].samp == NULL )
	    {
	      All[code].samp = (int *) malloc(dataSize*sizeof(int));
	      for ( i = 0; i < dataSize; i++ ) All[code].samp[i] = MeanCount0;
	    }

	  for ( i = 0; i < gridSize; i++ )
	    if ( Field[i] != All[code].missval ) All[code].samp[i+dataOffset]++;
	}
    }
}


void EchamAddRecord(int code, int gridID, int zaxisID, int levelID, int nmiss)
{
  static char func[] = "EchamAddRecord";
  int dataSize;
  int dataOffset;
  int nlevel;
  int gridSize;
  int gridtype, leveltype;

  gridtype   = gridInqType(gridID);
  leveltype  = zaxisInqType(zaxisID);
  nlevel     = zaxisInqSize(zaxisID);
  gridSize   = gridInqSize(gridID);
  dataSize   = gridSize*nlevel;
  dataOffset = gridSize*levelID;

  All[code].nmiss0 += nmiss;

  if ( gridtype == GRID_SPECTRAL )
    {
      /* ---------------------------------------------------------- */
      /* Found spectral field ! If needed, allocate memory and copy */
      /* ---------------------------------------------------------- */
      All[code].sfit = TRUE;
      All[code].hlev = nlevel;
      if ( nlevel > 1 && leveltype == ZAXIS_HYBRID )
	{
	  All[code].plev = NumLevelRequest;
	}
      else
	{
	  All[code].plev = 1;
	}

      if ( gridInqTrunc(gridID) != Truncation )
	Error(func, "Resolution error. Required %d - Found %d", Truncation, gridInqTrunc(gridID));

      if ( All[code].spectral0 == NULL )
	All[code].spectral0 = DoubleAlloc(dataSize, FieldName2(code, "spectral0"));

      RealCopy(All[code].spectral0+dataOffset, Field, gridSize);
    }
  else
    {
      All[code].hlev = nlevel;
      if ( nlevel > 1 && leveltype == ZAXIS_HYBRID && nlevel == NumLevel )
	{
	  All[code].plev = NumLevelRequest;
	  All[code].sfit = TRUE;
	}
      else
	{
	  All[code].plev = nlevel;
	  All[code].sfit = FALSE;
	}

      if ( Mean > 0 && (nmiss > 0 || All[code].samp) )
	{
	  int i;
	  if ( All[code].samp == NULL )
	    {
	      All[code].samp = (int *) malloc(dataSize*sizeof(int));
	      for ( i = 0; i < dataSize; i++ ) All[code].samp[i] = MeanCount0;
	    }

	  for ( i = 0; i < gridSize; i++ )
	    if ( Field[i] != All[code].missval ) All[code].samp[i+dataOffset]++;
	}

      if ( All[code].hybrid0 == NULL )
	All[code].hybrid0 = DoubleAlloc(dataSize, FieldName2(code, "hybrid0"));

      RealCopy(All[code].hybrid0+dataOffset, Field, gridSize);
    }
}


void printProcessStatus(int tsID)
{
  static int counthead = FALSE;

  if ( tsID == -1 )
    {
      if ( stdout_is_tty )
	{
	  fprintf(stdout, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
	  fflush(stdout);
	}

      counthead = FALSE;
    }
  else
    {
      if ( counthead == FALSE )
	{
	  if ( stdout_is_tty )
	    fprintf(stdout, " Process timestep :       ");

	  counthead = TRUE;
	}

      if ( stdout_is_tty )
	{
	  fprintf(stdout, "\b\b\b\b\b\b%6d", tsID);
	  fflush(stdout);
	}
    }
}


int setNextDate(void)
{
  int nrecs;
  int i;
  int righttime;
  int vdate, vtime;

  righttime = FALSE;
  while ( TRUE )
    {
      nrecs = streamInqTimestep(istreamID, TsID);
      if ( nrecs == 0 && ( Multi > 0 || Nfiles > 0 ) )
	{
	  printProcessStatus(-1);

	  SwitchFile();
	 
	  if ( istreamID >= 0 )
	    {
	      TsID = 0;
	      nrecs = streamInqTimestep(istreamID, TsID);
	    }
	}
      if ( nrecs == 0 ) break;

      vdate = taxisInqVdate(taxisID);
      vtime = taxisInqVtime(taxisID);

      setDateTime(&NextDate, vdate, vtime);

      for ( i = 0; i < nrqh; i++ )
	if ( hours[i] < 0 || hours[i] == NextDate.hr )
	  {
	    righttime = TRUE;
	    break;
	  }

      if ( righttime )
	break;
      else
	TsID += 1;	  
    }

  return (nrecs);
}


static int num_recs = 0;

void *readTimestep(void *arg)
{
  int i;
  int recID, varID, gridID, zaxisID, levelID, timeID;
  int level;
  int code, gridtype, leveltype;
  int levelOffset;
  int nmiss;
  int analysisData, nrecs;
  RARG *rarg = (RARG *) arg;

  nrecs        = rarg->nrecs;
  analysisData = rarg->lana;

  for ( code = 0; code < MaxCodes; code++ ) All[code].nmiss0 = 0;

  level = 0;
  levelOffset = 0;

  for ( recID = 0; recID < nrecs; recID++ )
    {
      streamInqRecord(istreamID, &varID, &levelID);

      code = vlistInqVarCode(ivlistID, varID);
      if ( code <= 0 || code >= MaxCodes ) continue;

      /* Skip records containing unneeded codes */

      if ( ! All[code].needed0 ) continue;

      vlistInqVar(ivlistID, varID, &gridID, &zaxisID, &timeID);

      leveltype = zaxisInqType(zaxisID);
	  
      /* Skip records with unselected levels */

      levelOffset = -1;
      /*
	if ( All[code].ozaxisID != All[code].izaxisID && ! Lhybrid2pressure )
      */
      if ( (All[code].ozaxisID != All[code].izaxisID) && (leveltype == ZAXIS_PRESSURE) )
	{
	  level = (int) zaxisInqLevel(zaxisID, levelID);
	  for ( i = 0; i < NumLevelRequest; ++i )
	    {
	      if ( LevelRequest[i] == level )
		{
		  levelOffset = i;
		  break;
		}
	    }

	  if ( levelOffset < 0 ) continue;

	  zaxisID = All[code].ozaxisID;
	  levelID = levelOffset;
	}

      if ( Debug )
	{
	  fprintf(stderr, "T%d", Truncation);

	  fprintf(stderr, "  Code %3d   Level%6d   %6.4d-%2.2d-%2.2d  %2.2d:%2.2d:00\n",
		  code, (int) zaxisInqLevel(zaxisID, levelID),
		  OldDate.yr, OldDate.mo, OldDate.dy, OldDate.hr, OldDate.mn);
	}

      streamReadRecord(istreamID, Field, &nmiss);

      if ( analysisData )
	AnalysisAddRecord(code, gridID, zaxisID, levelID, nmiss);
      else
	EchamAddRecord(code, gridID, zaxisID, levelID, nmiss);

      if ( iVertID != -1 && oVertID != -1 && (All[code].izaxisID == iVertID) )
	All[code].ozaxisID = oVertID;
    }

  TsID++;
  /*
    printf("%3d  date = %d  time = %04d\n", TsID, vdate, vtime);
  */
  num_recs = setNextDate();

  return ((void *) &num_recs);
}


void defineNextTimestep(void)
{
  static int otsID = 0;
  int vdate, vtime;

  vdate = getDate(OldDate);
  vtime = getTime(OldDate);
  taxisDefVdate(taxisID2, vdate);
  taxisDefVtime(taxisID2, vtime);

  /*
    printf("EOI %d %d %d %d  %d %d\n",  ostreamID, ostreamID2, otsID, MeanCount+1, vdate, vtime);
  */
  if ( Mean != 2 )
    {
      if ( otsID == 0 )
	{
	  vlistDefTaxis(ovlistID, taxisID2);
	  streamDefVlist(ostreamID, ovlistID);
	}
      taxisDefNumavg(taxisID2, MeanCount+1);
      streamDefTimestep(ostreamID,  otsID);
    }

  if ( Mean >= 2 )
    {
      if ( otsID == 0 )
	{
	  vlistDefTaxis(ovlistID2, taxisID2);
	  streamDefVlist(ostreamID2, ovlistID2);
	}
      taxisDefNumavg(taxisID2, MeanCount+1);
      streamDefTimestep(ostreamID2, otsID);
    }

  otsID++;
}


void setEndOfInterval(int nrecs)
{
  static char func[] = "setEndOfInterval";

  if ( nrecs == 0 )
    {
      EndOfInterval = TRUE;
    }
  else
    {
      if      ( OutputInterval == DAILY_INTERVAL )
	EndOfInterval = NewDate.dy != OldDate.dy;
      else if ( OutputInterval == MONTHLY_INTERVAL )
	EndOfInterval = NewDate.mo != OldDate.mo;
      else if ( OutputInterval == UNLIM_INTERVAL )
	EndOfInterval = FALSE;
      else
	Error(func, "output interval %d not implemented!\n", OutputInterval);
    }
}


void moveTimestep(void)
{
  int code;

  for ( code = 0; code < MaxCodes; code++ )
    All[code].nmiss = All[code].nmiss0;

  for ( code = 0; code < MaxCodes; code++ )
    if ( All[code].hybrid0 )
      {
	All[code].hybrid  = All[code].hybrid0;
	All[code].hybrid0 = NULL;
      }

  for ( code = 0; code < MaxCodes; code++ )
    if ( All[code].spectral0 )
      {
	All[code].spectral  = All[code].spectral0;
	All[code].spectral0 = NULL;
      }

  for ( code = 0; code < MaxCodes; code++ )
    if ( All[code].grid0 )
      {
	All[code].grid  = All[code].grid0;
	All[code].grid0 = NULL;
      }
}


void Control(int analysisData)
{
  static char func[] = "Control";
  int i;
  int righttime;
  int tsFirst, nrecs;
  int rdate, rtime;
  int vdate, vtime;
  int code;
  RARG rarg;
  void *statusp = NULL;
#if  defined  (HAVE_LIBPTHREAD)
  pthread_t thrID;
  pthread_attr_t attr;
  int rval;

  if ( ParallelRead )
    {
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    }
#endif

  for ( code = 0; code < MaxCodes; code++ )
    All[code].needed0 = All[code].needed;

  TsID = 0;

  righttime = FALSE;
  while ( (nrecs = streamInqTimestep(istreamID, TsID)) > 0 )
    {
      vdate = taxisInqVdate(taxisID);
      vtime = taxisInqVtime(taxisID);
      setDateTime(&StartDate, vdate, vtime);
      setDateTime(&NewDate, vdate, vtime);

      for ( i = 0; i < nrqh; i++ )
	if ( hours[i] < 0 || hours[i] == NewDate.hr )
	  {
	    righttime = TRUE;
	    break;
	  }

      if ( righttime )
	break;
      else
	TsID++;	  
    }

  if ( taxisInqType(taxisID) == TAXIS_RELATIVE )
    {
      rdate = taxisInqRdate(taxisID);
      rtime = taxisInqRtime(taxisID);
    }
  else
    {
      rdate = getDate(StartDate);
      rtime = getTime(StartDate);
    }

  if ( ofiletype == FILETYPE_NC || ofiletype == FILETYPE_NC2 || ofiletype == FILETYPE_NC4 )
    {
      taxisDefCalendar(taxisID2, CALENDAR_PROLEPTIC);
      taxisDefType(taxisID2, TAXIS_RELATIVE);
      taxisDefTunit(taxisID2, TUNIT_DAY);
      taxisDefRdate(taxisID2, rdate);
      taxisDefRtime(taxisID2, rtime);
    }

  OldDate = NewDate;

  tsFirst = TRUE;

  while ( nrecs > 0 )
    {
      rarg.nrecs = nrecs;
      rarg.lana  = analysisData;

      if ( tsFirst || ParallelRead == FALSE )
	{
	  if ( ParallelRead == FALSE )
	    {
	      statusp = readTimestep(&rarg);
	    }
#if  defined  (HAVE_LIBPTHREAD)
	  else
	    {
	      rval = pthread_create(&thrID, &attr, readTimestep, &rarg);
	      if ( rval != 0 ) Error(func, "pthread_create failed!");
	    }
#endif

	  if ( tsFirst )
	    {
	      if ( Type >  0 ) legini(Truncation, Latitudes);
	      if ( Type > 50 ) leginiOut(TruncationOut, LatitudesOut);
	    }

#if  defined  (HAVE_LIBPTHREAD)
	  if ( ParallelRead )
	    {
	      pthread_join(thrID, &statusp);
	      if ( *(int *)statusp < 0 )
		Error(func, "readTimestep error! (status = %d)", *(int *)statusp);
	    }
#endif
	  tsFirst = FALSE;
	}
#if  defined  (HAVE_LIBPTHREAD)
      else
	{
	  pthread_join(thrID, &statusp);
	  if ( *(int *)statusp < 0 )
	    Error(func, "readTimestep error! (status = %d)", *(int *)statusp);
	}
#endif
      nrecs = *(int *)statusp;

      MeanCount0 = MeanCount;
      NewDate = NextDate;

      moveTimestep();

#if  defined  (HAVE_LIBPTHREAD)
      if ( nrecs && ParallelRead )
	{
	  rval = pthread_create(&thrID, &attr, readTimestep, &rarg);
	  if ( rval != 0 ) Error(func, "pthread_create failed!");
	}
#endif

      setEndOfInterval(nrecs);
      
      printProcessStatus(TsID);

      if ( EndOfInterval ) printProcessStatus(-1);
      
      if ( Mean == 0 || EndOfInterval ) defineNextTimestep();

      if ( analysisData )
	processPL();
      else
	processML();

      PostProcess();
      
      if ( nrecs )
	{
	  if ( analysisData )
	    AnalysisDependencies(All, MaxCodes);
	  else
	    EchamDependencies(All, MaxCodes, Type, Source);
	}
	
      OldDate = NewDate;
    }

#if  defined  (HAVE_LIBPTHREAD)
  if ( ParallelRead )
    {
      pthread_attr_destroy(&attr);
    }
#endif
}


void setLevel(void)
{
  static char func[] = "setLevel";
  int k, l, found;
  int removeLevel[MaxLevel];
  double level;
  int checkLevel = TRUE;
  int numplevelDefault;  /* default pressure level */
  long plevelDefault[] = { 100000, 92500, 85000, 70000, 60000, 50000,  \
			    40000, 30000, 25000, 20000, 15000, 10000,  \
			     7000,  5000,  3000,  2000, 1000 };
  int numhlevelDefault;  /* default height level */
  long hlevelDefault[] = {  0, 1000, 2000, 5000, 10000, 15000, 20000, 25000, 30000 };

  numplevelDefault = sizeof(plevelDefault)/sizeof(plevelDefault[0]);
  numhlevelDefault = sizeof(hlevelDefault)/sizeof(hlevelDefault[0]);

  if ( iVertID != -1 )
    if ( zaxisInqType(iVertID) == ZAXIS_HYBRID && Type > 20 )
      Lhybrid2pressure = TRUE;

  lprintf(stdout);

  if ( NumLevelRequest == 0 )
    {
      if ( iVertID == -1 )
	{
	  fprintf(stdout," No level detected\n");
	}
      else
	{
	  if ( Lhybrid2pressure )
	    {
	      if ( unitsel == 0 )
		{
		  fprintf(stdout," Default pressure level selected:\n");
		  NumLevelRequest = numplevelDefault;
		  for ( l = 0; l < NumLevelRequest; l++ ) LevelRequest[l] = plevelDefault[l];
		  oVertID = zaxisCreate(ZAXIS_PRESSURE, NumLevelRequest);
		  zaxisDefLevels(oVertID, LevelRequest);
		}
	      else
		{
		  fprintf(stdout," Default height level selected:\n");
		  NumLevelRequest = numhlevelDefault;
		  for ( l = 0; l < NumLevelRequest; l++ ) LevelRequest[l] = hlevelDefault[l];
		  oVertID = zaxisCreate(ZAXIS_HEIGHT, NumLevelRequest);
		  zaxisDefLevels(oVertID, LevelRequest);
		}
	    }
	  else
	    {
	      if ( zaxisInqType(iVertID) == ZAXIS_HYBRID )
		fprintf(stdout," All detected hybrid level selected:\n");
	      else
		fprintf(stdout," All detected pressure level selected:\n");
	      NumLevelRequest = NumLevelFound;
	      for ( l = 0; l < NumLevelRequest; l++ ) LevelRequest[l] = LevelFound[l];
	      oVertID = iVertID;
	    }
	}
      checkLevel = FALSE;
    }
  else
    {
      if ( iVertID == -1 )
	{
	  fprintf(stdout," No level detected\n");
	  checkLevel = FALSE;
	}
      else if ( NumLevelRequest == 1 && LevelRequest[0] == 0 )
	{
	  fprintf(stdout," No level selected\n");
	  NumLevelRequest = 0;
	  checkLevel = FALSE;
	}
      else
	{
	  if ( Lhybrid2pressure )
	    {
	      if ( unitsel == 0 )
		fprintf(stdout," Selected pressure level:\n");
	      else
		fprintf(stdout," Selected height level:\n");
	    }
	  else
	    {
	      if ( zaxisInqType(iVertID) == ZAXIS_HYBRID )
		fprintf(stdout," Selected hybrid level:\n");
	      else
		{
		  if ( unitsel == 0 )
		    fprintf(stdout," Selected pressure level:\n");
		  else
		    fprintf(stdout," Selected height level:\n");
		}
	    }
	}
    }

  if ( iVertID != -1 )
    for ( l = 0; l < NumLevelRequest; l++ )
      fprintf(stdout, "  Level %2d = %13.4f\n", l+1, LevelRequest[l]);

  if ( checkLevel )
    {
      for ( k = 0; k < NumLevelRequest; k++ ) removeLevel[k] = FALSE;
      for ( k = 0; k < NumLevelRequest; k++ )
	{
	  level = LevelRequest[k];
	  for ( l = k+1; l < NumLevelRequest; l++ )
	    if ( level == LevelRequest[l] && removeLevel[l] == FALSE )
	      {
		fprintf(stdout, "  Level %2d = %13.4f double request\n",
			l+1, LevelRequest[l]);
		removeLevel[l] = TRUE;
	      }
	}

      l = 0;
      for ( k = 0; k < NumLevelRequest; k++ )
	if ( removeLevel[k] == FALSE )
	  LevelRequest[l++] = LevelRequest[k];

      NumLevelRequest = l;

      if ( AnalysisData || Type < 30 )
	{
	  for ( k = 0; k < NumLevelRequest; k++ ) removeLevel[k] = FALSE;
	  for ( k = 0; k < NumLevelRequest; k++ )
	    {
	      level = LevelRequest[k];
	      found = FALSE;
	      for ( l = 0; l < NumLevelFound; l++ )
		if ( level == LevelFound[l] ) found = TRUE;

	      if ( ! found )
		{
		  fprintf(stdout, "  Level %2d = %14.4f not in input\n",
			  k+1, LevelRequest[k]);
		  removeLevel[k] = TRUE;
		}
	    }

	  l = 0;
	  for ( k = 0; k < NumLevelRequest; k++ )
	    if ( removeLevel[k] == FALSE )
	      LevelRequest[l++] = LevelRequest[k];

	  if ( l != NumLevelRequest )
	    {
	      lprintf(stdout);
	      if ( labort )
		Error(func, "Inconsistent or invalid level list!");
	      else
		Warning(func, "Inconsistent or invalid level list!");
	    }

	  NumLevelRequest = l;
	}
    }

  lprintf(stdout);
}


void defineLevel(void)
{
  static char func[] = "defineLevel";
  int code, i;

  /* hybrid, pressure, height */

  switch ( Type ) {
  case  0:
  case 10:
  case 11:
  case 20:
    {
      if ( iVertID == -1 ) break;

      if ( zaxisInqType(iVertID) == ZAXIS_HYBRID )
	{
	  if ( oVertID == -1 )
	    {
	      if ( NumLevelRequest > NumLevelFound )
		Error(func, "Too much level requested");

	      if ( NumLevelFound == NumLevelRequest )
		{
		  for ( i = 0; i < NumLevelRequest; i++ )
		    if ( LevelRequest[i] != LevelFound[i] ) break;

		  if ( i == NumLevelRequest )
		    oVertID = iVertID;
		}

	      if ( oVertID == -1 && NumLevelRequest > 0 )
		{
		  oVertID = zaxisCreate(ZAXIS_HYBRID, NumLevelRequest);
		  zaxisDefLevels(oVertID, LevelRequest);
		  zaxisDefVct(oVertID, nvct, vct);
		}
	    }
	      
	  for ( code = 0; code < MaxCodes; code++ )
	    {
	      if ( All[code].selected )
		{
		  if ( All[code].izaxisID != -1 )
		    if ( zaxisInqType(All[code].izaxisID) == ZAXIS_HYBRID &&
			 zaxisInqSize(All[code].izaxisID) >= NumLevelRequest )
		      All[code].ozaxisID = oVertID;
		}
	    }
	}
      else
	Error(func, "%s level data unsupported for TYPE %d",
	      zaxisNamePtr(zaxisInqType(iVertID)), Type);	    

      break;
    }
  case 30:
  case 40:
  case 41:
  case 50:
  case 60:
  case 61:
  case 70:
    {
      if ( iVertID == -1 ) break;

      if ( oVertID == -1 )
	{
	  if ( unitsel == 0 )
	    oVertID = zaxisCreate(ZAXIS_PRESSURE, NumLevelRequest);
	  else
	    oVertID = zaxisCreate(ZAXIS_HEIGHT, NumLevelRequest);

	  zaxisDefLevels(oVertID, LevelRequest);
	}

      for ( code = 0; code < MaxCodes; code++ )
	{
	  if ( All[code].selected )
	    {
	      if ( All[code].izaxisID != -1 )
		if ( zaxisInqType(All[code].izaxisID) == zaxisInqType(iVertID) &&
		     zaxisInqSize(All[code].izaxisID) == NumLevel &&
		     zaxisInqSize(All[code].izaxisID) >  1 )
		  All[code].ozaxisID = oVertID;
	    }
	}

      break;
    }
  default:
    Error(func, "TYPE %d unsupported", Type);
  }
}

void
defineGrid(void)
{
  static char func[] = "defineGrid";
  int ogridID = -1;
  int code;

  /* spectral, fourier, gauss, zonal mean */

  switch ( Type ) {
  case  0:
  case 50:
    {
      if ( specGridID == -1 )
	{
	  if ( DimSP == 0 )
	    Error(func, "dim spectral undefined");
	  if ( Truncation == 0 )
	    Error(func, "truncation undefined");
	    
	  specGridID = gridCreate(GRID_SPECTRAL, DimSP);
	  gridDefTrunc(specGridID, Truncation);
	}

      ogridID = specGridID;
      break;
    }
  case 20:
  case 30:
  case 70:
    {
      if ( gaussGridID == -1 )
	{
	  if ( Longitudes == 0 )
	    Error(func, "number of longitudes undefined");
	  if ( Latitudes == 0 )
	    Error(func, "number of latitudes undefined");
	    
	  gaussGridID = gridCreate(GRID_GAUSSIAN, Longitudes*Latitudes);
	  gridDefXsize(gaussGridID, Longitudes);
	  gridDefYsize(gaussGridID, Latitudes);
	}

      ogridID = gaussGridID;
      break;
    }
  case 10:
  case 40:
  case 60:
    {
      if ( Fouriers == 0 )
	Error(func, "number of fourier coefficients undefined");
      if ( Latitudes == 0 )
	Error(func, "number of latitudes undefined");
	    
      ogridID = gridCreate(GRID_FOURIER, Fouriers*Latitudes);
      gridDefXsize(ogridID, Latitudes);
      gridDefYsize(ogridID, Fouriers);
      break;
    }
  case 11:
  case 41:
  case 61:
    {
      if ( Latitudes == 0 )
	Error(func, "Number of latitudes undefined");
	    
      ogridID = gridCreate(GRID_GAUSSIAN, Latitudes);
      gridDefXsize(ogridID, 1);
      gridDefYsize(ogridID, Latitudes);
      break;
    }
  default:
    Error(func, "TYPE %d unsupported", Type);
  }

  if ( ogridID != -1 )
    for ( code = 0; code < MaxCodes; code++ )
      {
	if ( All[code].selected )
	  {
	    All[code].ogridID = ogridID;
	  }
      }

  if ( ogridID == -1 )
    Error(func, "out grid undefined");
}

void setCodes(struct Control *All, int maxCodes, int numCodes)
{
  int code;
  int table, tableID;
  char *name, *longname;
  int varID;

  lprintf(stdout);

  if ( numCodes == 0 )
    {
      fprintf(stdout, " All detected codes selected:\n");
      
      for ( code = 0; code < maxCodes; code++ )
	if ( All[code].detected ) All[code].selected = 1;
    }
  else
    fprintf(stdout, " Selected codes:\n");
 
  fprintf(stdout, "  Table Code Name              Longname\n");
  fprintf(stdout, "  ----- ---- ----              --------\n");

  for ( code = 0; code < maxCodes; code++ )
    if ( All[code].selected )
      {
	table    = 0;
	name     = NULL;
	longname = NULL;
	varID    = All[code].ivarID;
	if ( varID == CDI_UNDEFID )
	  {
	    modelID  = vlistInqVarModel(ivlistID, 0);
	    table    = 128;
	    tableID  = tableInq(modelID, table, NULL);
	    All[code].tableID = tableID;
	  }
	else
	  {
	    tableID  = vlistInqVarTable(ivlistID, varID);
	    table    = tableInqNum(tableID);
	    name     = vlistInqVarNamePtr(ivlistID, varID);
	    longname = vlistInqVarLongnamePtr(ivlistID, varID);
	  }
	if ( name     == NULL )     name = tableInqParNamePtr(tableID, code);
	if ( longname == NULL ) longname = tableInqParLongnamePtr(tableID, code);

	fprintf(stdout, " %5d", table);
	fprintf(stdout, " %4d", code);
	if ( name == NULL )
	  fprintf(stdout, "  var%d", code);
	else
	  {
	    fprintf(stdout, "  %-16s", name);
	    if ( longname != NULL )
	      fprintf(stdout, "  %s", longname);
	  }
	fprintf(stdout, "\n");
      }
}

void checkNamelist(void)
{
  static char func[] = "checkNamelist";

  if ( Mean && Type < 20 )
    {
      Error(func, "Mean is only available for TYPE >= 20!");
    }

  if ( Extrapolate == FALSE && Type >= 30 )
    {
      if ( Type > 30 )
	Error(func, "EXTRAPOLATE = 0 is only available for TYPE = 30!");
      if ( Mean )
	Error(func, "EXTRAPOLATE = 0 is only available with MEAN = 0!");
    }
}

void parini(void)
{
  static char func[] = "parini";
  int i, c;
  long length;
  int head7dummy;
  int fileFormat, gribFormat, cdfFormat;
  int precision;
  char namelist[65536];

  if ( stdin_is_tty )
    fprintf(stdout, "\nEnter namelist parameter:\n");

  fseek(stdin, 0L, SEEK_END);
  length = ftell(stdin);
  if (length == 0L)
    {
      fprintf(stderr,"\n stdin not connected\n");
      Usage();
    }
  fseek(stdin,0L,SEEK_SET);
  i = 1;
  namelist[0] = ' ';
  c = getchar();
  while ((c != EOF) && i < (int)(sizeof(namelist) - 1))
    {
           if ((c >= '0' && c <= '9') ||
               (c == '-' || c == '.'))  namelist[i++] = c;
      else if  (c >= 'a' && c <= 'z')   namelist[i++] = c;
      else if  (c >= 'A' && c <= 'Z')   namelist[i++] = tolower(c);
      else c = ' ';

      if (c == ' ' && namelist[i-1] != ' ') namelist[i++] = c;
      c = getchar();
    }
  namelist[i] = 0;
  if ( Debug )
    {
      lprintf(stderr);
      fprintf(stderr,"  Length of namelist:%4d bytes\n",
	      (int) strlen(namelist));

      for (i = 0; i < (int)strlen(namelist); i += 60)
	fprintf(stderr,"  namelist[%02d]=%-60.60s\n", i, namelist+i);
      lprintf(stderr);
    }

  lprintf(stdout);
  fprintf(stdout, " Namelist:\n");

  Type           = scanpar(namelist, "type",  0);
  Multi          = scanpar(namelist, "multi", 0);
  Mean           = scanpar(namelist, "mean",  0);
  OutputInterval = scanpar(namelist, "interval", MONTHLY_INTERVAL);

  fileFormat     = scanpar(namelist, "format", 0);
  gribFormat     = scanparObsolate(namelist, "grib",   0);
  cdfFormat      = scanparObsolate(namelist, "netcdf", 0);

  if ( gribFormat && cdfFormat )
    Error(func, "GRIB or netCDF ?");

  switch ( fileFormat )
    {
    case 0: ofiletype = FILETYPE_SRV;  break;
    case 1: ofiletype = FILETYPE_GRB;  break;
    case 2: ofiletype = FILETYPE_NC;   break;
    case 3: ofiletype = FILETYPE_EXT;  break;
    case 4: ofiletype = FILETYPE_NC2;  break;
    case 6: ofiletype = FILETYPE_NC4;  break;
    default: Error(func, "unknown file format %d", fileFormat);
    }

  if ( gribFormat )  ofiletype = FILETYPE_GRB;
  if ( cdfFormat  )  ofiletype = FILETYPE_NC;

  precision = scanpar(namelist, "precision", 0);
  if ( precision )
    switch ( precision )
      {
      case  8: DataType = DATATYPE_PACK8;  break;
      case 16: DataType = DATATYPE_PACK16; break;
      case 24: DataType = DATATYPE_PACK24; break;
      case 32: DataType = DATATYPE_FLT32;  break;
      case 64: DataType = DATATYPE_FLT64;  break;
      default: Error(func, "unsupported data precision %d", precision);
      }


  head7dummy     = scanparObsolate(namelist, "head7",  0);
  unitsel        = scanpar(namelist, "unitsel",  0);
  DayIn          = scanpar(namelist, "dayinc",  30);
  Extrapolate    = scanpar(namelist, "extrapolate",  1);
  Szip           = scanpar(namelist, "szip",  0);
  TruncationOut  = scanparObsolate(namelist, "res",  0);
  LatitudesOut   = scanparObsolate(namelist, "lat",  0);
  mars           = scanparObsolate(namelist, "mars", 0);

  if ( Multi ) --Multi;

  if ( mars )
    {
      Grav         = C_MARS_GRAV;
      PlanetRadius = C_MARS_RADIUS;
      RD           = C_MARS_RD;
    }

  nrqh = scantime(namelist, hours, MaxHours);
  scancode(namelist, All, MaxCodes, &NumCodesRequest);

  scandarray(namelist, "level", LevelRequest, MaxLevel, &NumLevelRequest);
  if ( NumLevelRequest == 1 )
    if ( LevelRequest[0] == -1 )
      NumLevelRequest = 0;

  lprintf(stdout);

  checkNamelist();
}


static void dimcalc(void)
{
  if ( AnalysisData ) NumLevel = NumLevelRequest;
  if ( TruncationOut == 0 ) TruncationOut = Truncation;
  if ( TruncationOut > Truncation )
    fprintf(stderr,"*** Warning: Output Resolution > Data Resolution ! ***\n");

  if ( Latitudes == 0 )
    {
      Latitudes = 2 * ((Truncation*3 + 3) / 4);
      if ( Truncation == 30 ) Latitudes = 48;
    }

  if ( LatitudesOut == 0 ) LatitudesOut = Latitudes;

  if ( Longitudes == 0 )
    {
      Longitudes = Latitudes * 2;
      if ( Truncation == 62 ) Longitudes = 192;
    }

  if ( LongitudesOut == 0 )
    {
      LongitudesOut = LatitudesOut * 2;
      if ( TruncationOut == 62 ) LongitudesOut = 192;
    }

  Waves         = Truncation + 1;
  Fouriers      = Waves * 2;
  DimSP         = (Truncation + 1) * (Truncation + 2);
  DimFC         = Latitudes * Fouriers;
  DimGP         = Latitudes * Longitudes;
  Dim3GP        = NumLevel * DimGP;
  Dim3FC        = NumLevel * DimFC;
  Dim3SP        = NumLevel * DimSP;
  HalfLevels    = NumLevel + 1;
  WavesOut      = TruncationOut + 1;
  FouriersOut   = WavesOut * 2;
  DimSPOut      = (TruncationOut + 1) * (TruncationOut + 2);
  DimFCOut      = LatitudesOut * FouriersOut;
  DimGPOut      = LatitudesOut * LongitudesOut;
  Dim3GPOut     = NumLevel * DimGPOut;
  Dim3FCOut     = NumLevel * DimFCOut;
  Dim3SPOut     = NumLevel * DimSPOut;
  DimSPOut_half = DimSPOut / 2;


  if ( AnalysisData )
    fprintf(stdout, " Found Ana or Re-Ana Data\n");

  fprintf(stdout, " Dimensions:\n");

  fprintf(stdout, "  Truncation        = %4d\n", Truncation);
  fprintf(stdout, "  Levels            = %4d\n", NumLevel);
  fprintf(stdout, "  Latitudes         = %4d\n", Latitudes);
  fprintf(stdout, "  Longitudes        = %4d\n", Longitudes);
  if ( TruncationOut != Truncation )
    {
      fprintf(stdout, "  Output Truncation = %4d\n", TruncationOut);
      fprintf(stdout, "  Output Latitudes  = %4d\n", LatitudesOut);
      fprintf(stdout, "  Output Longitudes = %4d\n", LongitudesOut);
    }
  lprintf(stdout);
}


/* ----------------------------------------------------------- */
/* Extract basic dimension information                         */
/* ----------------------------------------------------------- */

void precntl(void)
{
  static char func[] = "precntl";
  int l;
  int code = 0;
  int nvars, ngrids, nverts, ntsteps;
  int gridID, zaxisID, varID, timeID;
  int i, index, leveltype, gridtype;
  int datasize, numlevel;
  int instID, modelID;
  int vertfound = 0;
  int nhzaxis = 0;


  nvars   = vlistNvars(ivlistID);
  ngrids  = vlistNgrids(ivlistID);
  nverts  = vlistNzaxis(ivlistID);
  ntsteps = vlistNtsteps(ivlistID);

  instID = vlistInqInstitut(ivlistID);

  if ( Debug )
    {
      Message(func, "nvars      = %d", nvars);
      Message(func, "ngrids     = %d", ngrids);
      Message(func, "nverts     = %d", nverts);
      Message(func, "ntsteps    = %d", ntsteps);
    }

  for ( index = 0; index < ngrids; index++ )
    {
      gridID   = vlistGrid(ivlistID, index);
      gridtype = gridInqType(gridID);
      datasize = gridInqSize(gridID);

      if ( datasize > FieldDim ) FieldDim = datasize;

      if ( gridtype == GRID_SPECTRAL && Truncation == 0 )
	{
	  specGridID = gridID;
	  Truncation = gridInqTrunc(gridID);
	}
      else if ( gridtype == GRID_GAUSSIAN && Latitudes == 0 )
	{
	  gaussGridID = gridID;
	  Longitudes  = gridInqXsize(gridID);
	  Latitudes   = gridInqYsize(gridID);
	}
      else
	Error(func, "%s data unsupported!\n", gridNamePtr(gridtype));
    }

  if ( Truncation == 0 )
    {
      if ( Latitudes )
	{
	  switch ( Latitudes ) {
	  case 512: Truncation = 511; break;
	  case 320: Truncation = 213; break;
	  case 192: Truncation = 127; break;
	  case 160: Truncation = 106; break;
	  case 128: Truncation =  85; break;
	  case  96: Truncation =  63; break;
	  case  94: Truncation =  62; break;
	  case  64: Truncation =  42; break;
	  case  48: Truncation =  31; break;
	  case  32: Truncation =  21; break;
	  default :
	    fprintf(stderr,"%d Gaussian latitudes not supported.\n", Latitudes);
	  }
	}
    }

  for ( index = 0; index < nverts; index++ )
    {
      zaxisID   = vlistZaxis(ivlistID, index);
      leveltype = zaxisInqType(zaxisID);
      numlevel  = zaxisInqSize(zaxisID);
      /*
	printf("leveltype : %d %d\n", leveltype, zaxisInqSize(zaxisID));
      */	
      if ( numlevel > 1 )
	{
	  if ( leveltype == ZAXIS_HYBRID || leveltype == ZAXIS_PRESSURE )
	    {
	      if ( leveltype == ZAXIS_HYBRID && Vctfile == NULL )
		{
		  nhzaxis++;
		  if ( numlevel != (zaxisInqVctSize(zaxisID)/2 - 1) )
		    {
		      if ( ! (numlevel == 191 && zaxisInqVctSize(zaxisID) == 0) )
			{
			  Warning(func, "Skip %d hybrid level data with %d levels!",
				  (zaxisInqVctSize(zaxisID)/2 - 1), numlevel);
			  continue;
			}
		    }
		}

	      if ( iVertID != - 1 )
		Warning(func, "More than %d different vertical grid structure found!", vertfound);

	      vertfound++;

	      if ( iVertID != -1 ) continue;

	      iVertID = zaxisID;
	      NumLevelFound = numlevel;
	      LevelFound = (double *) malloc(NumLevelFound*sizeof(double));
	      for ( l = 0; l < NumLevelFound; l++ )
		LevelFound[l] = (int) zaxisInqLevel(zaxisID, l);

	      if ( leveltype == ZAXIS_HYBRID )
		{
		  if ( Vctfile == NULL )
		    {
		      if ( zaxisInqVctSize(zaxisID) )
			{
			  nvct = zaxisInqVctSize(zaxisID);

			  if ( vct == NULL )
			    {
			      vct = (double *) malloc(nvct*sizeof(double));
			      memcpy(vct, zaxisInqVctPtr(zaxisID), nvct*sizeof(double));
			    }
			}
		      else
			{
			  if ( numlevel == 191 )
			    {
			      fprintf(stderr," Using internal VCT for L191\n");
			      nvct = (191+1)*2;
			      vct = (double *) malloc(nvct*sizeof(double));
			      memcpy(vct, VCT_L191, nvct*sizeof(double));
			      zaxisDefVct(zaxisID, nvct, vct);
			    }
			  else
			    {
			      Error(func, "VCT not defined in inputfile!");
			    }
			}
		    }

		  if ( numlevel != (nvct/2 - 1) )
		    Error(func, "Number of hybrid levels %d does not match vct levels %d",
			  numlevel, nvct/2-1);

		  if ( Debug )
		    for ( i = 0; i < nvct/2; i++ )
		      fprintf(stderr," vct: %4d %10.4f %10.4f\n", i, vct[i], vct[i+nvct/2]);
		}

	      if ( leveltype == ZAXIS_PRESSURE ) AnalysisData = TRUE;
	    }
	}
    }

  if ( nhzaxis > 0 && nvct == 0 ) Error(func, "VCT missing!");

  NumLevel = NumLevelFound;

  if (  specGridID != -1 ) Spectral = TRUE;
  if ( gaussGridID != -1 ) Gaussian = TRUE;

  if ( Debug )
    fprintf(stderr, "   T = %3d   L = %2d\n", Truncation, NumLevelFound);

  if ( Debug )
    fprintf(stderr," CODE CHECK\n");

  instID  = vlistInqVarInstitut(ivlistID, 0);
  modelID = vlistInqVarModel(ivlistID, 0);

  lprintf(stdout);
  fprintf(stdout, " Institute : ");
  if ( instID == CDI_UNDEFID )
    fprintf(stdout, "unknown\n");
  else
    {
      if ( institutInqLongnamePtr(instID) )
	fprintf(stdout, "%s\n", institutInqLongnamePtr(instID));
      else
	fprintf(stdout, "name unknown\n");
    }

  fprintf(stdout, " Source    : ");
  if ( modelID == CDI_UNDEFID )
    fprintf(stdout, "unknown\n");
  else
    {
      if ( modelInqNamePtr(modelID) )
	{
	  if ( strncmp(modelInqNamePtr(modelID), "ECHAM5", 6) == 0 ) Source = S_ECHAM5;
	  fprintf(stdout, "%s\n", modelInqNamePtr(modelID));
	}
      else
	fprintf(stdout, "name unknown\n");
    }

  for ( varID = 0; varID < nvars; varID++ )
    {
      vlistInqVar(ivlistID, varID, &gridID, &zaxisID, &timeID);
      code      = vlistInqVarCode(ivlistID, varID);
      if ( code <= 0 || code >= MaxCodes )
	{
	  Warning(func, "Code number %d out of range, variable ignored!", code);
	  continue;
	}
      gridtype  = gridInqType(gridID);
      numlevel  = zaxisInqSize(zaxisID);
      leveltype = zaxisInqType(zaxisID);

      All[code].ivarID  = varID;
      All[code].igridID = gridID;
      All[code].ogridID = gridID;
      All[code].izaxisID = zaxisID;
      All[code].ozaxisID = zaxisID;

      All[code].detected = TRUE;

      if ( Debug )
	fprintf(stderr,"Code %3d  Levels = %3d  LevelType = %3d  GridType = %3d\n",
		code, numlevel, leveltype, gridtype);
    }

  if ( Debug )
    Message(func, "FieldDim = %d\n", FieldDim);

  Field = (double *) malloc(FieldDim*sizeof(double));

  if ( Debug )
    for ( code = 0; code < MaxCodes; code++ )
      {
	if ( All[code].detected )
	  fprintf(stderr," Detected Code %3d with %3d level\n",
		  code, zaxisInqSize(All[code].izaxisID));
      }
}

/*
 * -----------------------------------------------------------
 * Define output variables
 * -----------------------------------------------------------
 */

void postcntl(void)
{
  int code = 0;
  int gridID, zaxisID;
  int ovarID, ogridID, ozaxisID;
  int ovarID2;
  int ivarID, instID, modelID, tableID;
  char *name, *longname, *units;
  char histstring[99];
  int testmode = 0;
  int datatype;

  sprintf(histstring, "afterburner version %s  type = %d", VERSION, Type);

  gmInqHistory(istreamID);
  if ( Mean != 2 ) gmDefHistory(ostreamID, histstring);
  if ( Mean >= 2 ) gmDefHistory(ostreamID2, histstring);

  if ( Debug || testmode ) lprintf(stdout);
  if ( Debug || testmode )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].detected )
	{
	  gridID = All[code].igridID;
	  zaxisID = All[code].izaxisID;
	  fprintf(stderr," Detected Code %3d  grid %-8s size %5d  level %2d %-8s\n",
		  code, gridNamePtr(gridInqType(gridID)), gridInqSize(gridID),
		  zaxisInqSize(zaxisID), zaxisNamePtr(zaxisInqType(zaxisID)));
	}


  if ( Debug || testmode ) lprintf(stdout);
  if ( Debug || testmode )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].needed )
	{
	  fprintf(stderr,"   Needed Code %3d\n", code);
	}

  for ( code = 0; code < MaxCodes; code++ )
    if ( All[code].selected )
      {
	name     = NULL;
	longname = NULL;
	units    = NULL;
	ivarID   = All[code].ivarID;
	ogridID  = All[code].ogridID;
	ozaxisID = All[code].ozaxisID;

	if ( ogridID == -1 )
	  {
	    /*
	    Warning(func, "undefined grid for code %d", code);
	    */
	    continue;
	  }
	if ( ozaxisID == -1 )
	  {
	    /*
	    Warning(func, "undefined level for code %d", code);
	    */
	    continue;
	  }

	instID   = vlistInqVarInstitut(ivlistID, ivarID);
	modelID  = vlistInqVarModel(ivlistID, ivarID);
	tableID  = vlistInqVarTable(ivlistID, ivarID);

	All[code].missval  = vlistInqVarMissval(ivlistID, ivarID);
	All[code].samp     = NULL;

	if ( DataType != -1 )
	  datatype = DataType;
	else
	  datatype = vlistInqVarDatatype(ivlistID, ivarID);

	if ( All[code].comp )
	  {
	    tableID = All[code].tableID;
	  }
	else
	  {
	    name     = vlistInqVarNamePtr(ivlistID, ivarID);
	    longname = vlistInqVarLongnamePtr(ivlistID, ivarID);
	    units    = vlistInqVarUnitsPtr(ivlistID, ivarID);
	  }

	if ( Mean != 2 )
	  {
	    vlistDefTaxis(ovlistID, taxisID2);
	    ovarID = vlistDefVar(ovlistID, ogridID, ozaxisID, TIME_VARIABLE);
	    vlistDefVarCode(ovlistID, ovarID, code);
	    All[code].ovarID = ovarID;
	    vlistDefVarInstitut(ovlistID, ovarID, instID);
	    vlistDefVarModel(ovlistID, ovarID, modelID);
	    vlistDefVarTable(ovlistID, ovarID, tableID);
	    if ( Mean )     vlistDefVarTimave(ovlistID, ovarID, 1);
	    if ( name )     vlistDefVarName(ovlistID, ovarID, name);
	    if ( longname ) vlistDefVarLongname(ovlistID, ovarID, longname);
	    if ( units )    vlistDefVarUnits(ovlistID, ovarID, units);
	    vlistDefVarDatatype(ovlistID, ovarID, datatype);
	    vlistDefVarMissval(ovlistID, ovarID, All[code].missval);
	  }

	if ( Mean >= 2 )
	  {
	    vlistDefTaxis(ovlistID2, taxisID2);
	    ovarID2 = vlistDefVar(ovlistID2, ogridID, ozaxisID, TIME_VARIABLE);
	    vlistDefVarCode(ovlistID2, ovarID2, code);
	    All[code].ovarID2 = ovarID2;
	    vlistDefVarInstitut(ovlistID2, ovarID2, instID);
	    vlistDefVarModel(ovlistID2, ovarID2, modelID);
	    vlistDefVarTable(ovlistID2, ovarID2, tableID);
	    if ( Mean )     vlistDefVarTimave(ovlistID2, ovarID2, 1);
	    if ( name )     vlistDefVarName(ovlistID2, ovarID2, name);
	    if ( longname ) vlistDefVarLongname(ovlistID2, ovarID2, longname);
	    if ( units )    vlistDefVarUnits(ovlistID2, ovarID2, units);
	    vlistDefVarDatatype(ovlistID2, ovarID2, datatype);
	    vlistDefVarMissval(ovlistID2, ovarID2, All[code].missval);
	  }
      }

  if ( Debug || testmode ) lprintf(stdout);
  if ( Debug || testmode )
    for ( code = 0; code < MaxCodes; code++ )
      if ( All[code].selected )
	{
	  gridID  = All[code].ogridID;
	  zaxisID = All[code].ozaxisID;
	  fprintf(stderr," Selected Code %3d  grid %-8s size %5d  level %2d %-8s\n",
		  code, gridNamePtr(gridInqType(gridID)), gridInqSize(gridID),
		  zaxisInqSize(zaxisID), zaxisNamePtr(zaxisInqType(zaxisID)));
	}
}


void Usage(void)
{
  fprintf(stderr, "\nafter [options] <InputFiles> <OutputFile> <VarianceFile>\n");
#if defined (_OPENMP)
  fprintf(stderr, "     option -P <nthreads> : Set number of OpenMP threads\n");
#endif
  fprintf(stderr, "     option -a            : Forces analysis data process\n");
  fprintf(stderr, "     option -c            : Print available codes and names\n");
  fprintf(stderr, "     option -d            : Debug mode\n");
  fprintf(stderr, "     option -v <vctfile>  : Read vct from vctfile\n");
  /*  fprintf(stderr, "     option -h : help (this output)\n"); */
  /*  fprintf(stderr, "     option -p : parallel read on\n"); */
  fprintf(stderr, "  <InputFiles> : ECHAM or ECMWF Ana or ReAna files\n");
  fprintf(stderr, "  <OutputFile> : GRIB, netCDF or SERVICE format file\n");
  fprintf(stderr, "<VarianceFile> : GRIB, netCDF or SERVICE format file\n");
  fprintf(stderr, "  namelist is read from <stdin>\n");
  fprintf(stderr, "  output is written to <stdout>\n\n");

  fprintf(stderr, "  dedault Namelist: \n");
  fprintf(stderr, "  &SELECT\n");
  fprintf(stderr, "    TYPE = 0, CODE = -1, LEVEL = -1, MULTI = 0, DAYIN = 30,\n");
  fprintf(stderr, "    MEAN = 0, TIMESEL = -1, UNITSEL = 0,\n");
  fprintf(stderr, "    FORMAT = 0, PRECISION = 0, SZIP = 0\n");
  fprintf(stderr, "  &END\n");

  exit(1);
}

static
void readVct(const char *vctfile)
{
  static char func[] = "readVct";
  FILE *fp;
  char line[1024];
  int i, n;
  double va, vb;

  fp = fopen(vctfile, "r");
  if ( fp == NULL ) SysError(func, "Open failed on %s", vctfile);

  while ( fgets(line, 1023, fp) ) nvct++;

  nvct *= 2;
  vct = (double *) malloc(nvct*sizeof(double));

  rewind(fp);
  for ( i = 0; i < nvct/2; i++ )
    {
      fgets(line, 1023, fp);
      sscanf(line, "%d %lg %lg", &n, &va, &vb);
      vct[i]        = va;
      vct[i+nvct/2] = vb;
    }
  fprintf(stdout, "  Reading VCT for %d hybrid levels from file %s\n", nvct/2-1, vctfile);

  fclose(fp);
}

static
void version(void)
{
#if defined (COMPILER)
  fprintf(stderr, "Compiler: %s\n", COMPILER);
#endif
#if defined (COMP_VERSION)
  fprintf(stderr, " version: %s\n", COMP_VERSION);
#endif
#if defined (HAVE_LIBSZ) || defined (_OPENMP)
  fprintf(stderr, "    with:");
#if defined (HAVE_LIBSZ)
  fprintf(stderr, " libsz");
#endif
#if defined (_OPENMP)
  fprintf(stderr, " OpenMP");
#endif
  fprintf(stderr, "\n");
#endif
#if defined (USER_NAME) && defined(HOST_NAME) && defined(SYSTEM_TYPE)
  fprintf(stderr, "Compiled: by %s on %s (%s) %s %s\n",
	  USER_NAME, HOST_NAME, SYSTEM_TYPE, __DATE__, __TIME__);
#endif
  cdiPrintVersion();
  fprintf(stderr, "\n");
}


extern char *optarg;
extern int optind, opterr, optopt;

int main(int argc, char *argv[])
{
  static char func[] = "after";
  int   i, code;
  char *proc = argv[0];
  char  Line[132];
  int c;
  int fargc0, fargcn;
  FILE *fp;
  int numThreads = 0;
  extern int dmemory_ExitOnError;

  dmemory_ExitOnError = 1;

  setStartTime();

  { /* check character device on stdin and stdout */
    struct stat statbuf;
    fstat(0, &statbuf);
    if ( S_ISCHR(statbuf.st_mode) ) stdin_is_tty = 1;  
    fstat(1, &statbuf);
    if ( S_ISCHR(statbuf.st_mode) ) stdout_is_tty = 1;  
  }

  /* ------------------- */
  /*  print information  */
  /* ------------------- */

  lprintf(stdout);
  fprintf(stdout,"  afterburner version %s\n", VERSION);
  fprintf(stdout,"  ECHAM & analyses postprocessor        \n");

  if ( sizeof(double) != 8 || sizeof(int) < 4 )
    {
      fprintf(stderr, "byte size of type double %d\n", (int) sizeof(double));
      fprintf(stderr, "byte size of type int %d\n",    (int) sizeof(int));
      fprintf(stderr, "byte size of type size_t %d\n", (int) sizeof(size_t));
      return(1);
    }

  fp = fopen("/pf/m/m214003/doc/afterburner.doc","r");
  if ( fp )
    {
      do
	{
	  fgets(Line, 130, fp);
	  fprintf(stdout, "%s", &Line[1]);
	}
      while ( ! feof(fp) && Line[0] == '#' );
      fclose(fp);
    }

  /* --------------------- */
  /*  options & filenames  */
  /* --------------------- */

  while ( (c = getopt(argc, argv, "P:b:v:acdgmpVw")) != EOF )
    switch (c)
      {
      case 'a': AnalysisData = 1; break;
      case 'b': Message(func, "option -b not longer needed!\n\n"); break;
      case 'c': printCodes(); break;
      case 'd': Debug = 1; break;
      case 'g': ECMWFMars = TRUE; break;
      case 'p': ParallelRead = TRUE; break;
      case 'P': numThreads = atoi(optarg); break;
      case 'V': version(); break;
      case 'v': Vctfile = optarg; break;
      case 'w': labort = FALSE; break;
      default:  Message(func, "option -%c unsupported\n", c); Usage();
      }

#if defined (_OPENMP)
  /* ParallelRead = TRUE; */

  lprintf(stdout);
  if ( numThreads <= 0 ) numThreads = 1;
  omp_set_num_threads(numThreads);
  if ( omp_get_max_threads() > omp_get_num_procs() )
    fprintf(stdout, " Number of threads is greater than number of CPUs=%d!\n", omp_get_num_procs());
  fprintf(stdout, " OpenMP:  num_procs = %d  max_threads = %d\n",
	  omp_get_num_procs(), omp_get_max_threads());
#else
  if ( numThreads > 0 )
    {
      fprintf(stderr, "Option -P failed, OpenMP support not compiled in!\n");
      return(-1);
    }
#endif

  if ( ParallelRead )
    {
#if  defined  (HAVE_LIBPTHREAD)
      fprintf(stdout, " Parallel read enabled\n");
#else
      fprintf(stdout, " Parallel read disabled\n");
      ParallelRead = FALSE;
#endif
    }

  for ( code = 0; code < MaxCodes; code++ )
    {
      All[code].ivarID   = -1;
      All[code].ovarID   = -1;
      All[code].ovarID2  = -1;
      All[code].izaxisID = -1;
      All[code].ozaxisID = -1;
      All[code].igridID  = -1;
      All[code].ogridID  = -1;
      All[code].tableID  = -1;
    }

  fargc0 = optind;
  fargcn = argc;

  if ( optind < argc ) ifile = argv[optind++];
  if ( ! ifile )
    {
      Message(func, "*** Missing input file ***");
      Usage();
    }

  parini(); /* read namelist parameter */

  fprintf(stdout, "   Input File: %-25s\n", ifile);
  if ( Mean >= 2 )
    {
      if ( fargcn-fargc0 >= 3 ) ofile2 = argv[--fargcn];
      if ( ! ofile2 )
	{
	  Message(func, "*** Missing variance file ***");
	  Usage();
	}
    }

  if ( Mean != 2 )
    {
      if ( optind < argc ) ofile = argv[optind++];
      if ( fargcn-fargc0 >= 2 ) ofile = argv[--fargcn];
      if ( ! ofile )
	{
	  Message(func, "*** Missing output file ***");
	  Usage();
	}
      fprintf(stdout, "  Output File: %-25s\n", ofile);
    }

  Nfiles = fargcn-fargc0-1;
  if ( Nfiles > 0 )
    {
      if ( Multi > 0 )
	Error(func, "Namelist parameter MULTI works only with one inputfile");

      ifiles = (char **) malloc(Nfiles*sizeof(char*));
      for ( i = 0; i < Nfiles; i++ )
	ifiles[i] = argv[--fargcn];
    }

  if ( ofile2 )
    fprintf(stdout, "Variance File: %-25s\n", ofile2);

  if ( Debug )
    {
      fprintf(stderr, "* Debug on!                              *\n");
      fprintf(stderr, "  Maximum ffts to run in parallel:  %d\n", get_nfft());
    }

  /* read option VCT */
  if ( Vctfile ) readVct(Vctfile);

  /* --------------------- */
  /*  open in/output file  */
  /* --------------------- */

  cdiDefGlobal("REGULARGRID", 1); 

  istreamID = streamOpenRead(ifile);
  if ( istreamID < 0 ) cdiError(istreamID, "Open failed on %s", ifile);

  ivlistID = streamInqVlist(istreamID);
  taxisID  = vlistInqTaxis(ivlistID);
  taxisID2 = taxisDuplicate(taxisID);

  ifiletype = streamInqFiletype(istreamID);

  if ( Mean != 2 )
    {
      ostreamID  = streamOpenWrite(ofile,  ofiletype);
      if ( ostreamID < 0 ) cdiError(ostreamID, "Open failed on %s", ofile);

      if ( Szip ) streamDefZtype(ostreamID, COMPRESS_SZIP);

      ovlistID = vlistCreate();
    }
  if ( Mean >= 2 )
    {
      ostreamID2 = streamOpenWrite(ofile2, ofiletype);
      if ( ostreamID2 < 0 ) cdiError(ostreamID2, "Open failed on %s", ofile2);

      if ( Szip ) streamDefZtype(ostreamID, COMPRESS_SZIP);

      ovlistID2 = vlistCreate();
    }

  /* ---------------- */
  /*  pre-processing  */
  /* ---------------- */
  precntl();

  /* ----------------- */
  /*  initializations  */
  /* ----------------- */

  setCodes(All, MaxCodes, NumCodesRequest);

  if ( unitsel == 2 )
    for (i = 0; i < NumLevelRequest; i++) LevelRequest[i] = LevelRequest[i] * 1000;

  if ( ! AnalysisData )
    for (i = 0; i < NumLevelRequest; i++)
      {
	if ( (LevelRequest[i] >= 65535) && unitsel && ofiletype == FILETYPE_GRB )
	  {
	    fprintf(stderr,"\n Level %9.2f out of range (max=65535)!\n", LevelRequest[i]);
	    exit(1);
	  }

	if ( (LevelRequest[i] == 0) && !unitsel && Type >= 20 && NumLevelRequest > 1 )
	  {
	    fprintf(stderr,"\n Level %9.2f illegal for Type %d\n", LevelRequest[i], Type);
	    exit(1);
	  }
      }

  setLevel();

  dimcalc();

  rcoslat          = (double *) malloc(Latitudes*sizeof(double));
  coslat           = (double *) malloc(Latitudes*sizeof(double));
  rcoslatOut       = (double *) malloc(Latitudes*sizeof(double));
  coslatOut        = (double *) malloc(Latitudes*sizeof(double));
  DerivationFactor = (double *) malloc(Latitudes*sizeof(double));

  if ( Type < 50 && AnalysisData )
    {
      fprintf(stderr," ::::::::::::::::::::::::::::::::::::::::::::::\n");
      fprintf(stderr," -> Type < 50 is not appropriate for Analysis.\n");
      fprintf(stderr," -> Please check wether you can use Type >= 50.\n");
      fprintf(stderr," -> Premature Exit. Sorry.\n");
      exit(1);
    }

  if ( Type == 10 || Type == 40 || Type == 60 )
    {
      if ( ofiletype == FILETYPE_GRB )
	Error(func, "Can't write fourier coefficients to GRIB!");
      else if ( ofiletype == FILETYPE_NC || ofiletype == FILETYPE_NC2 ||
		ofiletype == FILETYPE_NC4 )
	Error(func, "Can't write fourier coefficients to netCDF!");
    }

  if ( TruncationOut < Truncation )
    {
      if ( Type < 50 )
	{
	  fprintf(stderr," ::::::::::::::::::::::::::::::::::::::::::::::\n");
	  fprintf(stderr," -> Parameter res:\n");
	  fprintf(stderr," -> Please avoid Type < 50 for technical reasons.\n");
	  fprintf(stderr," -> Premature Exit. Sorry.\n");
	  exit(1);
	}
      if ( Gaussian )
	{
	  fprintf(stderr," ::::::::::::::::::::::::::::::::::::::::::::::\n");
	  fprintf(stderr," -> Parameter res:\n");
	  fprintf(stderr," -> Please avoid 'downmelting' with gaussian\n");
	  fprintf(stderr," -> grids. No source code yet !\n");
	  fprintf(stderr," -> Premature Exit. Sorry.\n");
	  exit(1);
	}
    }

  if ( TruncationOut > Truncation )
    {
      fprintf(stderr," ::::::::::::::::::::::::::::::::::::::::::::::\n");
      fprintf(stderr," -> Parameter res\n");
      fprintf(stderr," -> Please avoid OutputTruncation > Truncation!\n");
      fprintf(stderr," -> No source code yet !\n");
      fprintf(stderr," -> Premature Exit. Sorry.\n");
      exit(1);
    }

  filename = strrchr(ifile,'/');
  if (filename == 0) filename = ifile;
  else               filename++ ;

  experimentID = atoi(filename);

  if ( Type >= 30 && Type < 50 &&
      (All[DIVERGENCE].selected || All[VELOPOT].selected ||
       All[VORTICITY].selected  || All[STREAM].selected  ||
       AnalysisData) )
    {
      /*
      int newtype = 0;
      */
      if ( Type == 30 ) Type = 70;
      if ( Type == 40 ) Type = 60;
      if ( Type == 41 ) Type = 61;

      if ( AnalysisData )
	fprintf(stderr,"\n TYPE changed to %d (for analysis data)\n", Type);
      else
	fprintf(stderr,"\n TYPE changed to %d (with code %d, %d, %d or %d)\n",
		Type, DIVERGENCE, VELOPOT, VORTICITY, STREAM);
      /*
      if ( Type == 30 ) newtype = 70;
      if ( Type == 40 ) newtype = 60;
      if ( Type == 41 ) newtype = 61;

      if ( AnalysisData )
	fprintf(stderr,"\n Attention: TYPE isn't changed to %d anymore (for analysis data)!!!\n", Type);
      else
	fprintf(stderr,"\n Attention: TYPE isn't changed to %d anymore (with code %d, %d, %d or %d)!!!\n",
		newtype, DIVERGENCE, VELOPOT, VORTICITY, STREAM);
      */
    }

  if ( AnalysisData ) AnalysisDependencies(All, MaxCodes);
  else
    {
      EchamDependencies(All, MaxCodes, Type, Source);
      All[GEOPOTENTIAL].needed |= Type >= 30 || All[SLP].comp || All[GEOPOTHEIGHT].comp;
    }

  /*  if ( All[U_WIND].needed || All[V_WIND].needed ) */
  if ( All[U_WIND].comp || All[V_WIND].comp )
    {
      dv2uv_f1 = DoubleAlloc(DimSPOut_half, "dv2uv_f1");
      dv2uv_f2 = DoubleAlloc(DimSPOut_half, "dv2uv_f2");
      geninx(TruncationOut, dv2uv_f1, dv2uv_f2);
    }

  /* --------- */
  /*  Control  */
  /* --------- */

  defineLevel();

  defineGrid();

  postcntl(); /* define output variables */

  Control(AnalysisData);

  if ( ostreamID2 != CDI_UNDEFID ) streamClose(ostreamID2);
  if ( ostreamID  != CDI_UNDEFID ) streamClose(ostreamID);
  streamClose(istreamID);

  if ( rcoslat )          free(rcoslat);
  if ( coslat )           free(coslat);
  if ( rcoslatOut )       free(rcoslatOut);
  if ( coslatOut )        free(coslatOut);
  if ( DerivationFactor ) free(DerivationFactor);

  if ( Field ) free(Field);

  if ( poli ) free(poli);
  if ( pold ) free(pold);
  if ( pdev ) free(pdev);
  if ( pol2 ) free(pol2);
  if ( pol3 ) free(pol3);

  FreeMean(All);

  procstat(proc, Truncation);

  return(0);
}
